/*
 * inout.h - functions for reading and writing files
 *
 * These functions are intended for piping a file from stdin to stdout with
 * the following additions:
 *
 * - Whole input lines are always kept in memory. When an error occurs, we can
 *   print the whole line and line number.
 * - When trying to read the next character, you don't just get the character
 *   back, but a pointer to the next characters, at least until the next
 *   newline or EOF character. So you can read characters without actually
 *   removing them from the input queue.
 * - Output is buffered first; only a call to write_flush () actually writes
 *   the chars to stdout.
 * - We can insert special code characters (labeled by an arbitrary integer)
 *   into the output (they never occur in the input). When flushed, these
 *   codes disappear. Before flushing, they can be replaced later by an
 *   arbitrary string.
 * - There are no limits on the lengths of the input lines or the output
 *   buffer.
 *
 */

#ifndef __INOUT_H__
#define __INOUT_H__

#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <assert.h>

using namespace std;



/* If we include a file defining INOUT_DEBUG, we print to stderr what's going
 * on in the output queue.
 *
 */

#include "config.h"



char *inout_init (); // initialize and return pointer to first input char
char *write_next (); // Write current input char to output, and go to next
		     // input char
void write_char (char); // Write a constant char to the output
void write_string (char *); // same for a string
char *write_white (); // Copy whitespace, including newlines and comments, to
		      // output.
char *skip_next ();  // Do not write current input char to output, go to next
		     // input char
void write_code (int); // Write a (non-printable) code character to the output.
int write_code (); // same, but generate a new un-used code that is then
		   // returned.
void replace_code (int, char*); // Replace a non-printable code (that has not
				// been flushed yet) with a string
void write_flush (); // Flush the output to stdout
void stop_error (char *); // Terminate the program with an error message,
			  // print full line with number in which the error
			  // occurred

#endif

