/*
 * relinv.cc - compute GW splittings in relative theory
 *
 */

#include "relinv.h"

/*
 * If we include a file setting RELINV_DEBUG to a number, next() will print
 * out every combination that it returns, provided that the current level
 * does not exceed RELINV_DEBUG.
 *
 */

#include "config.h"



// The constructor to make a relinv from an inv.

relinv::relinv (inv& iv, int lvln, int skipc)
{
	int i,j;

	// These are the _maximum_ possible values
	ncls = iv.points; // the number of different classes in iv
	ncpt = iv.cdeg+1; // the internal component can have degree 0,
			  // the others must have degree at least 1.
	points = iv.points;
	lvl = lvln;
	skip = 0;

	// Allocate memory for the various arrays
	if (ncls) assert (cls = new pdata[ncls+3]);
	else cls = NULL;
	assert (cls_vdim = new int[ncls+3]);
	assert (cpt_vdim = new int[ncpt]);
	assert (cpt_tm = new int[ncpt]);
	assert (cpt_prim = new int[ncpt]);
	assert (value_ptr = new int[(ncls+3)*ncpt]);
	assert (same_ptr = new int[(ncls+5)*ncpt]);
	assert (node_ptr = new int[3*(iv.cgenus+1)*ncpt]);
	firstmax = ncpt;

	// Read the classes.
	cpt_prim[0] = 0;
	ncls = 3; // The first three entries are for degree, genus, and
		  // loop nodes
	for (i=0; i<iv.points; i++) {
		if (i==skipc) {
			skip = 1;
			skipclass = iv.pt[i];
		} else if (!i || iv.pt[i]!=iv.pt[i-1]) {
			cls[ncls] = iv.pt[i];
			cls_vdim[ncls] = -iv.pt[i].vdim(iv.xdim);
			value(0,ncls++) = 1;
			if (iv.pt[i].Hpow<cpt_prim[0])
				cpt_prim[0] = iv.pt[i].Hpow;
		} else {
			value(0,ncls-1)++;
		}
	}
	// ncls is set correctly now and will not change any more.

	// At the beginning, we have only the internal component, of maximal
	// degree and genus, and with all the classes. Set the unused
	// variables to their starting values.
	for (i=1; i<ncpt; i++) for (j=0; j<ncls; j++) value(i,j) = 0;

	ncpt = 1;
	cdeg(0) = iv.cdeg;
	cgenus(0) = iv.cgenus;
	cloop(0) = 0;

	// Set the virtual dimension of a genus change of 1.
	cls_vdim[1] = 2*(2-iv.xdim);

	// Set same
	for (i=0; i<ncls; i++) same(0,i) = 1;

	// Set the coefficient
	coeff = 1;

	// Set the number of primitive classes. Note that the topological
	// Euler class is equal to the top Chern class of the tangent bundle,
	// and at the same time (xdim+1) + (-1)^(xdim) times the number of
	// primitive classes.
	if (iv.xdeg!=1) {
		prims = (pow(1-iv.xdeg,iv.xdim+2)+iv.xdeg-1) / iv.xdeg;
		if (iv.xdim & 1) prims = -prims;
	} else prims = 0;

	totalgenus = iv.cgenus;
	xdim = iv.xdim;
	xdeg = iv.xdeg;

	// Compute the minimum H class to put on the nodes. This is 0 except
	// in g=0 if there are no points, and the degree of the hypersurface
	// is small enough.
	node_minclass = 0;
	if (!totalgenus && !iv.points) {
		if ( xdeg==xdim+2 ) node_minclass = xdim-1; // Calabi-Yau
		if ( xdeg==xdim+1 ) node_minclass = xdim; // Fano index 1
		if ( xdeg<xdim+1 ) node_minclass = xdim+1; // Fano index >1
	}

	// Are there moving classes?
	noclass = 0;
	if (!iv.points) noclass = 1;
	if (iv.points==1 && skip) noclass = 1;

	#ifdef RELINV_DEBUG
	if (lvl<=RELINV_DEBUG && mult())
		cout << "*relinv* level " << lvl << ": " << *this << endl;
	#endif
}



/*
 * Go to the next degree splitting. Splitting of the degree is different
 * from splitting of genus and the classes, for two reasons:
 * - There is no previous 'same' data to be read.
 * - The number of components can change. It's limited by the restriction
 *   that the degree of all components except the first one must be positive.
 *
 * Note that the first component is always special (it contains the
 * central component). The splittings are enumerated in the following
 * order (example: total degree 7)
 *  7 |
 *  6 | 1
 *  5 | 2
 *  5 | 1,1
 *  4 | 3
 *  4 | 2,1
 *  4 | 1,1,1
 *  3 | 4
 *  3 | 3,1
 *  3 | 2,2
 *  3 | 2,1,1
 *  3 | 1,1,1,1
 *  2 | 5
 *  ...
 * The general rule: find the rightmost number > 1, decrease it by 1,
 * and split the remaining degree onto the next components, always
 * taking the maximum possible degree (i.e. the same as the decreased
 * one, except for a possible remainder in the last one).
 *
 */

bool relinv::deg_next ()
{
	int i,j,k,d;

	// Look for the rightmost degree >1.
	for (i=ncpt-1; i>0; i--) if ( cdeg(i)>1 ) {
		// Found the rightmost degree > 1.
		//
		// We're about to change this value. If the entries left of
		// this one are the same, we have to decrease the corresponding
		// 'same' number by 1.
		j = i;
		if ( --j>0 && cdeg(j)==cdeg(i) ) {
			// We have more components of this degree. Find the
			// leftmost one and decrease 'same'.
			while ( j>1 && cdeg(j-1)==cdeg(i) ) j--;
			same(j,0)--;
			// Promote this same value to deeper class levels.
			for (k=1; k<ncls; k++) same(j,k) = same(j,0);
		}

		// Decrease the current degree by 1.
		cdeg(i)--;
		cpt_vdim[i] -= 2*(xdim+2);
		cpt_tm[i] = 0;

		// Now split up the remaining degree.
		d = ncpt-i; // the remaining degree
		k = same(i,0) = d/cdeg(i) + 1; // this is how often we can have
					       // this value.
		// Promote this same value to deeper class levels:
		for (j=1; j<ncls; j++) same(i,j) = same(i,0);
		i++;

		// Write this degree the appropriate number of times.
		for (j=2; j<=k; j++) {
			d -= (cdeg(i) = cdeg(i-1));
			cpt_vdim[i] = 2*((2+xdim)*cdeg(i)+xdim-2); // note that
				// the genus is 0 when we change the degree
			cpt_tm[i] = 0;
			i++;
		}

		// If there is a fractional part left, write it, with 'same'
		// number 1.
		if (d) {
			cdeg(i) = d;
			cpt_vdim[i] = 2*((2+xdim)*d+xdim-2);
			cpt_tm[i] = 0;
			same(i,0) = 1;
			// Promote 'same' value:
			for (k=1; k<ncls; k++) same(i,k) = same(i,0);
			i++;
		}

		// Set the new number of components:
		ncpt = i;
		return 1;
	}

	// All degrees of the external components are 1. Decrease the degree
	// of the internal component.
	if ( !cdeg(0) ) return 0; // done
	cdeg(0)--;
	cdeg(1) = ncpt;
	cpt_vdim[1] = 2*((2+xdim)*ncpt+xdim-2);
	cpt_tm[1] = 0;
	ncpt = 2;
	for (k=0; k<ncls; k++) same(1,k) = 1;

	return 1;
}



/*
 * Go to the next splitting within a sequence of 'same' classes.
 *
 * We pass an array index (number of component and number of class). By
 * reading the 'same' value in the previous level, we know how many components
 * starting from here are considered to be the same for the current class
 * level. Among these components and in this class level, go to the next
 * splitting, or reset the splitting to the initial value and return 0.
 *
 * This is very similar to the degree splitting in deg_next, except that:
 * - the value 0 is allowed,
 * - no components are added; rather there is a fixed number of components
 *   onto which we have to split the classes.
 *
 * For example, to split 6 classes onto 3 same components, we'll count as
 * follows:
 *
 * 6,0,0
 * 5,1,0
 * 4,2,0
 * 4,1,1
 * 3,3,0
 * 3,2,1  A
 * 2,2,2  B
 *
 * Note that we cannot just decrease the rightmost 1 entry as before: we have
 * to check that the remaining space is enough to distribute the remaining
 * classes. In the above example, from A to B we decrease 3 to 2 rather than
 * 2 to 1, because (3,1,1,1) doesn't fit onto 3 components.
 *
 * This function is used for the genus and the classes splitting. If we split
 * classes, we update coeff according to the multinomial coefficient.
 * Relatively speaking, we have a factor of 1/c! in coeff for every entry c.
 *
 */

bool relinv::same_next (int m, int n)
{
	int c,i,j,k,l,N;

	if (!value(m,n)) return 0; // If there are no classes, there is
				   // nothing to split.

	N = same(m,n-1); // the current 'same' value. We can split the
			 // classes onto the next N components.

	if (N==1) return 0; // If same is 1, there's nothing to split.

	i = m+N-1; // The rightmost component within the 'same' range.
	c = value(i,n); // The number of classes read, from right to left.
	for (j=2; j<=N; j++) {
		c += (k = value(--i,n));
		if (n!=1) coeff *= factorial (value(i+1,n));
		cpt_vdim[i+1] -= value(i+1,n)*cls_vdim[n];
		if (n!=1) cpt_tm[i+1] -= value(i+1,n)*cls[n].mult;
		value(i+1,n) = 0;
		if ( k>1 && (c+k-2)/(k-1)<=j ) goto FOUND; // -> OK, can
			// decrease k at position i, and the classes will fit.
	}

	// No more splittings. Reset values and return 0.
	if (n!=1) coeff *= factorial (value(i,n)) / factorial (c);
	cpt_vdim[i] += (c-value(i,n))*cls_vdim[n];
	if (n!=1) cpt_tm[i] += (c-value(i,n))*cls[n].mult;
	value(i,n) = c;
	for (k=n; k<ncls; k++) {
		same(i,k) = 1;
		same(i+1,k) = N-1;
	}
	return 0;

FOUND:	// We can decrease value(i,n) and distribute the remaining classes
	// onto the right adjacent components. Do so.

	// We're about to decrease the current value. If there are same values
	// left of this one, decrease the corresponding 'same' value.
	for (k=0; k<N-j; k++) if (value(i-k-1,n)!=value(i,n)) break;
	if (k) {
		same(i-k,n)--;
		// Promote this same value to deeper class levels.
		for (l=n+1; l<ncls; l++) same(i-k,l) = same(i-k,n);
	}

	// Decrease the current value.
	if (n!=1) coeff *= value(i,n);
	cpt_vdim[i] -= cls_vdim[n];
	if (n!=1) cpt_tm[i] -= cls[n].mult;
	value(i,n)--;

	// Write the corresponding 'same' value.
	k = same(i,n) = c/value(i,n);
	c -= value(i,n);
	for (l=n+1; l<ncls; l++) same(i,l) = k;

	// Write the current value as often as needed.
	for (l=1; l<k; l++) {
		if (n!=1) coeff /= factorial (value(i,n));
		cpt_vdim[i+1] += value(i,n)*cls_vdim[n];
		if (n!=1) cpt_tm[i+1] += value(i,n)*cls[n].mult;
		value(i+1,n) = value(i,n);
		i++;
		j--;
		c -= value(i,n);
	}

	// Write the remaining classes.
	if (c) {
		j--;
		if (n!=1) coeff /= factorial (c);
		cpt_vdim[++i] += c*cls_vdim[n];
		if (n!=1) cpt_tm[i] += c*cls[n].mult;
		value (i,n) = c;
		for (l=n; l<ncls; l++) same(i,l) = 1;
	}

	if (--j>0) {
		// There are still j zeroes to write.
		i++;
		for (l=n; l<ncls; l++) same(i,l) = j;
		for (l=1; l<=j; l++) value(i++,n) = 0;
	}

	return 1;
}



/*
 * Go to the next splitting distributing classes onto 'same' groups.
 *
 * We pass a number of the class. We assume that in every 'same' block of
 * this class, all classes are in the first component. We go to the next
 * splitting that distributes the classes onto the blocks in a different
 * way. After the call, again all classes will be in the first components
 * of 'same' blocks.
 *
 * If there are no more splittings, reset the splitting to the initial value
 * (everything in component 0) and return 0.
 *
 * For example, to split 5 classes onto 3 'same' groups, we'll count as
 * follows:
 *
 * 5,0,0
 * 4,1,0
 * 3,2,0
 * 2,3,0
 * 1,4,0
 * 0,5,0
 * 4,0,1
 * 3,1,1
 * 2,2,1
 * 1,3,1
 * 0,4,1
 * 3,0,2
 * ...
 *
 * In other words, the strategy is as follows:
 * - find the leftmost non-zero entry,
 * - from this entry, move one to the group to the right...
 * - ... and the rest to the very left.
 *
 * This function is used for the genus and the classes splitting. If we split
 * classes, we update coeff according to the multinomial coefficient.
 * Relatively speaking, we have a factor of 1/c! in coeff for every entry c.
 *
 */

bool relinv::group_next (int n)
{
	int i;
	int m = 0; // the starting component
	int m2;

	// Find the leftmost non-zero entry. If all entries are 0, return.
	while (!value(m,n)) if ( (m += same(m,n-1)) >= ncpt ) return 0;

	if ( (m2=m+same(m,n-1)) >= ncpt ) {
		// End of loop, restore initial state.
		if (!m) return 0;
		value(0,n) = value(m,n);
		value(m,n) = 0;
		cpt_vdim[m] -= value(0,n)*cls_vdim[n];
		cpt_vdim[0] += value(0,n)*cls_vdim[n];
		if (n!=1) {
			cpt_tm[m] -= value(0,n)*cls[n].mult;
			cpt_tm[0] += value(0,n)*cls[n].mult;
		}
		for (i=n; i<ncls; i++) same(m,i) = same(m,n-1);
		return 0;
	}

	// Move one class to the right.
	if ( !(value(m2,n)++) ) {
		for (i=n; i<ncls; i++) same(m2,i) = 1;
		if ( same(m2,n-1)>1 )
			for (i=n; i<ncls; i++) same(m2+1,i) = same(m2,n-1)-1;
	}
	if (n!=1) {
		coeff *= value(m,n);
		coeff /= value(m2,n);
		cpt_tm[m] -= cls[n].mult;
		cpt_tm[m2] += cls[n].mult;
	}
	value(m,n)--;
	cpt_vdim[m] -= cls_vdim[n];
	cpt_vdim[m2] += cls_vdim[n];

	// Move the remaining classes to the very left, if we're not already
	// there.
	if (m && value(m,n)) {
		value(0,n) = value(m,n);
		value(m,n) = 0;
		cpt_vdim[m] -= value(0,n)*cls_vdim[n];
		if (n!=1) cpt_tm[m] -= value(0,n)*cls[n].mult;
	}

	// Update 'same' in the current position.
	if (!value(m,n)) for (i=n; i<ncls; i++) same(m,i) = same(m,n-1);

	return 1;
}



/*
 * Go to the next 'node configuration'.
 *
 * This is not really a splitting, as there is no fixed total number that has
 * to be split up. Rather, any component of genus g can have between 0 and
 * g 'loop nodes'. If it has n loop nodes, the effective genus g of the
 * component comes from a 'true genus' g-n of the component, and n loops.
 *
 * Going to the next configuration is very similar to 'counting'; we raise
 * the first number up to the maximum value, and after that we reset it and
 * raise the next one. Here is an example where the first two numbers belong
 * to a "same" block, and the maximum entry is 3:
 *
 * 0,0 | 0
 * 1,0 | 0
 * 2,0 | 0
 * 3,0 | 0
 * 1,1 | 0
 * 2,1 | 0
 * 3,1 | 0
 * 2,2 | 0
 * 3,2 | 0
 * 3,3 | 0
 * 0,0 | 1
 * 1,0 | 1
 * ...
 *
 * Note that the first component (the internal one) cannot have any loop
 * nodes.
 *
 */

bool relinv::loop_next ()
{
	int i,j,k,l;

	// Raise numbers, starting from the first component. Continue as
	// long as we get an 'overflow'.
	j = 1; // the beginning of the current 'same' block
	k = 1+same(1,1); // the beginning of the next 'same' block
	for (i=1; i<ncpt; i++) { // Skip the internal component
		if (i>=k) { // we moved to the next 'same' block
			j = k;
			k += same(i,1);
		}
		if (++cloop(i) <= cgenus(i)) goto FOUND;
		// Overflow in this component...
		cpt_vdim[i] += (cloop(i)-1)*cls_vdim[1];
		cloop(i) = 0;
	}

	// Overflow in every component; end of loop.
	return 0;

FOUND:	cpt_vdim[i] -= cls_vdim[1];
	// Set previous values in the current 'same' block to the same number.
	for (k=j; k<i; k++) {
		cpt_vdim[k] -= (cloop(k)-cloop(i))*cls_vdim[1];
		cloop(k) = cloop(i);
	}
	if ( same(i,2)>1 )
		for (l=2; l<ncls; l++) same(i+1,l) = same(i,2)-1;
	for (l=2; l<ncls; l++) same (j,l) = i-j+1;
	return 1;
}



/*
 * Check if the node configuration at component ci is OK. This means:
 * - If primitive classes do not occur in pairs, return 0.
 * - If the dimension condition of the invariant would require the node
 *   multiplicities to be half-integers, negative, or the total multiplicity
 *   to be bigger than the intersection product, return 0.
 * Otherwise, return 1 and set the node multiplicities to their initial value.
 *
 */

bool relinv::nodemult_init (int ci)
{
	int i,j,k,l,vd;

	// See the special implementation at the beginning of nodeclass_init:
	if (noclass && !cloop(ci)) return 1;

	// Check that primitive classes occur in pairs.
	if (xdeg>1) { // otherwise there are no primitive classes anyway

		// If there are any primitive classes at marked points other
		// than nodes, check these first.
		j = 0; // The primitive classes that we checked so far.
		if (cpt_prim[0]) {
			k = 0;
			for (i=3; i<ncls; i++) if (cls[i].Hpow<j) {
				// We have found a primitive class that we
				// haven't checked yet.
				j = cls[i].Hpow;
				k = 0; // how often does it occur?
				// Check the non-node points...
				for (l=i; l<ncls; l++) if (cls[l].Hpow==j)
					k += value(ci,l);
				// ... and finally the nodes.
				for (l=0; l<=cloop(ci); l++)
					if (node_class(ci,l)==j) k++;
				// Found the power. If it's odd, return 0.
				if (k&1) return 0;
			}
		}

		// Now check the primitive classes that occur at the nodes
		// only.
		for (i=0; i<=cloop(ci); i++) if (node_class(ci,i)<j) {
			j = node_class(ci,i);
			k = 1;
			for (l=i+1; l<=cloop(i); l++)
				if (node_class(ci,l)==j) k++;
			// Found the power. If it's odd, return 0.
			if (k&1) return 0;
		}
		// OK, all primitive classes occur in pairs.
	}

	// Now compute the total multiplicity needed at the nodes.
	// Compute the virtual dimension so far.
	vd = cpt_vdim[ci];
	for (j=0; j<=cloop(ci); j++) vd -=
		(node_class(ci,j)<0 ? xdim : 2*node_class(ci,j));
	// We have already checked that primitive classes occur in pairs.
	// So the virtual dimension must be even.
	vd = vd/2+cloop(ci)+1; // the total multiplicity needed
	if (cpt_tm[ci]+vd>xdeg*cdeg(ci)) return 0; // too big total mult.

	// Set the first multiplicity to the maximum, the others to 1.
	if ( (node_mult(ci,0)=vd-cloop(ci)) < 1 ) return 0;
	for (j=1; j<=cloop(ci); j++) node_mult(ci,j) = 1;

	return 1;
}



/*
 * Go to the next multiplicity splitting in component ci. Return 0 if there
 * are no further splittings. The splitting is basically the same as the
 * genus splitting above (a combination of same_next and group_next).
 *
 */

bool relinv::nodemult_next (int ci)
{
	int ni = 0; // node number
	int c,i,j,k,N;

	// There's no freedom in the multiplicities if there is only one node:
	if (!cloop(ci)) return 0;

	// First try to change the splitting within node_same blocks.
	while (ni<=cloop(ci)) {
		N = node_same (ci,ni); // the current node_same number
		if (N==1) goto NXTSAM;

		c = node_mult(ci,ni+N-1)-1; // sum of the multiplicities read
		for (i=ni+N-2; i>=ni; i--) {
			c += (k=node_mult (ci,i))-1;
			node_mult (ci,i+1) = 1;
			if (k>2 && (c+(k-3))/(k-2)<=ni+N-i)
				goto FOUND; // OK, can decrease multiplicity
		}

		// No more splittings here. Reset values and try the next
		// node_same group.
		node_mult(ci,ni) = c+1;
		goto NXTSAM;

FOUND:		node_mult(ci,i)--; // decrease multiplicity
		c -= (--k)-1;
		while (c>=k-1) {
			node_mult (ci,++i) = k;
			c -= (k-1);
		}
		if (c) node_mult (ci,++i) = c+1;
		return 1;

NXTSAM:		ni += N;
	}

	// We cannot change the distribution within the current node_same
	// blocks. All node_same blocks have been reset to their original
	// state. Change the splitting of the multiplicities onto these
	// blocks.
	ni = 0;

	// Find the leftmost entry that is not 1.
	while (node_mult(ci,ni)==1)
		if ( (ni+=node_same(ci,ni)) > cloop(ci) ) return 0;

	// We want to move one multiplicity to the node_same block on the
	// right. If there is no more block to the right, return.
	if ( (i=ni+node_same(ci,ni))>cloop(ci) ) return 0;

	node_mult(ci,i)++;
	node_mult(ci,ni)--;

	// Move the remaining multiplicity to the very left, if we're not
	// already there.
	if (!ni) return 1;
	node_mult(ci,0) += node_mult(ci,ni)-1;
	node_mult(ci,ni) = 1;
	return 1;
}



/*
 * Initialize the node classes at component ci to their initial values. It
 * is not guaranteed that there is a node multiplicity configuration for this
 * class configuration that leads to a non-zero invariant. Use nodemult_init
 * to figure that out. Return 0 only if it's very obvious that not even a
 * sensible node class configuration can be found.
 *
 */

bool relinv::nodeclass_init (int ci)
{
	int i,j;

	// We provide a special fast function if there are no external classes
	// and the component has only one node. In this case, we also set
	// the multiplicity already (and nodemult_init does nothing any more).
	if (noclass && !cloop(ci)) {
		// We cannot have primitive classes. There is only one
		// cohomology class and one multiplicity. The smallest
		// class is obtained by setting the multiplicity to the
		// maximum value. We cannot have primitive classes.
		j = xdeg*cdeg(ci);
		i = cpt_vdim[ci]/2-(j-1);
		if (i>xdim) return 0;
		if (i<0) {
			j += i;
			i = 0;
			if (j<1) return 0;
		}
		node_class (ci,0) = i;
		node_mult (ci,0) = j;
		node_same (ci,0) = 1;
		return 1;
	}

	// Set all nodes to the minimal class
	for (j=0; j<=cloop(ci); j++)
		node_class(ci,j) = node_minclass;
	node_same(ci,0) = cloop(ci)+1;

	return 1;
}



/*
 * Go to the next node class configuration at component ci. It is not
 * guaranteed that there is a node multiplicity configuration for this class
 * configuration that leads to a non-zero invariant. Use nodemult_init to
 * figure that out. Return 0 if there are no more splittings.
 *
 */

bool relinv::nodeclass_next (int ci)
{
	int i,ni;

	// We provide a special fast function if there are no external classes
	// and the component has only one node. Also, we set the multiplicity
	// already (and nodemult_next does nothing any more).
	if (noclass && !cloop(ci)) {
		if (node_class(ci,0)==xdim || node_mult(ci,0)==1) return 0;
		node_class (ci,0)++;
		node_mult (ci,0)--;
		return 1;
	}

	for (ni=0; ni<=cloop(ci); ni++) {
		// See if we can go to another class at node ni of
		// component ci.

		// If there are still bigger H powers, multiply with H.
		if (node_class(ci,ni)>=0 && node_class(ci,ni)<xdim) {
			node_class(ci,ni)++;
			goto RAISED;
		}

		// H powers done. Can we switch to a primitive class? We
		// do _not_ need this if
		// - the degree of the variety is 1, or
		// - the dimension of a primitive class is too big for
		//   node_minclass, or
		// - there are no primitive classes to start with, and there
		//   is only one node on this component.
		if (node_class(ci,ni)==xdim && xdeg>1
				&& xdim>=2*node_minclass
				&& (cloop(ci) || cpt_prim[0])) {
			node_class(ci,ni) = -1;
			goto RAISED;
		}

		// Can we move on to a next primitive class? We can go one
		// further than what we have in the next component. We can
		// also go one further than what we have on the next node
		// on this component.
		if (node_class(ci,ni)<0 &&
				(node_class(ci,ni)>=cpt_prim[ci] ||
				 node_same(ci,ni)>1) ) {
			// yes, can go to the next primitive class. We can
			// at least go one further than in the next
			// component. We can also go one further than
			// the next node in the same component.
			node_class(ci,ni)--;
			goto RAISED;
		}
	}

	// No node at component ci can be raised further.
	return 0;

RAISED:	// We have just gone to the next class at node ni of component ci.
	// Update the node_same value of the next node to reflect
	// that we have changed the current one.
	if ( node_same(ci,ni)>1 )
		node_same(ci,ni+1) = node_same(ci,ni)-1;
	// Set the previous nodes of the same component to
	// the same value, and update the node_same value at the
	// first node accordingly.
	for (i=0; i<ni; i++)
		node_class(ci,i) = node_class(ci,ni);
	node_same (ci,0) = ni+1;

	return 1;
}



/*
 * Initialize node classes and multiplicities to the first useable value,
 * or return 0 if this doesn't exist. This is a combination of
 * nodeclass_init and nodemult_init, called for every component.
 *
 */

bool relinv::node_init ()
{
	int i=1;
	int j,k;

	while (i<ncpt) {
		// Try to initialize the current component i.
		if (!nodeclass_init(i)) return 0;
		while (!nodemult_init(i)) if (!nodeclass_next(i)) return 0;

		// OK. Copy to this same block.
		for (j=i+1; j<i+same(i,ncls-1); j++) {
			for (k=0; k<=cloop(j); k++) {
				node_class(j,k) = node_class(i,k);
				node_mult(j,k) = node_mult(i,k);
				node_same(j,k) = node_same(i,k);
			}
		}

		// Next same block
		i += (same(i,ncls+1) = same(i,ncls) = same(i,ncls-1));
	}

	return 1;
}



/*
 * Go to the next configuration of node classes and multiplicities. Return
 * 0 if there is no further configuration. This is basically a combination
 * of nodeclass_next and nodemult_next, keeping track of the 'same' values.
 *
 */

bool relinv::node_next ()
{
	int ci; // the component we're looking at
	int cj; // the beginning of the current same block
	int i,j;

	// Try to go to the next multiplicity configuration first.
	ci = cj = 1;
	while (ci<ncpt) {
		if (nodemult_next(ci)) {
			// OK, next mult configuration at component ci.
			// Write it to the 'same' block.
			for (i=cj; i<ci; i++) for (j=0; j<=cloop(i); j++)
				node_mult (i,j) = node_mult (ci,j);
			// Update 'same'.
			if (same (ci,ncls+1) > 1)
				same (ci+1,ncls+1) = same (ci,ncls+1)-1;
			same (cj,ncls+1) = ci-cj+1;
			return 1;
		}

		// Advance to next component.
		if (++ci==cj+same(cj,ncls)) {
			// We're passing over a 'same' block. Reset the block.
			nodemult_init(cj);
			for (i=cj+1; i<ci; i++) for (j=0; j<=cloop(i); j++)
				node_mult (i,j) = node_mult (cj,j);
			same (cj,ncls+1) = same (cj,ncls);
			cj = ci;
		}
	}

	// There is no next multiplicity configuration. Try to go to the next
	// class configuration.
	ci = cj = 1;

	while (ci<ncpt) {
		while (nodeclass_next(ci)) if (nodemult_init(ci)) {
			// OK, next class configuration at component ci
			// with a corresponding set of multiplicities.
			// Write it to the 'same' block.
			for (i=cj; i<ci; i++) for (j=0; j<=cloop(i); j++) {
				node_mult (i,j) = node_mult (ci,j);
				node_class (i,j) = node_class (ci,j);
				node_same (i,j) = node_same (ci,j);
			}
			// Update 'same'.
			if (same (ci,ncls) > 1)
				same (ci+1,ncls+1) = same (ci+1,ncls)
					= same (ci,ncls)-1;
			same (cj,ncls+1) = same (cj,ncls) = ci-cj+1;
			// Update cpt_prim.
			for (i=ci-1; i>=1; i--) {
				cpt_prim[i] = cpt_prim[i+1];
				if (node_class(i+1,0)<cpt_prim[i])
					cpt_prim[i] = node_class(i+1,0);
			}
			return 1;
		}
		
		// Advance to next component.
		if (++ci==cj+same(cj,ncls-1)) {
			// We're passing over a 'same' block. Reset the block.
			if (!nodeclass_init(cj)) return 0;
			while (!nodemult_init(cj))
				if (!nodeclass_next(cj)) return 0;
			for (i=cj+1; i<ci; i++) for (j=0; j<=cloop(i); j++) {
				node_mult (i,j) = node_mult (cj,j);
				node_class (i,j) = node_class (cj,j);
				node_same (i,j) = node_same (cj,j);
			}
			same (cj,ncls+1) = same (cj,ncls) = same (cj,ncls-1);
			cj = ci;
		}
	}

	return 0;
}



/*
 * Go to the next relative splitting. Return 0 if there are no more splittings.
 *
 * This just calls all the previous functions in the right order.
 *
 * If the function exits with value 1, it is guaranteed that the external
 * invariants satisfy the dimension condition, and that they do not have too
 * big total multiplicity. It is also guaranteed that they contain primitive
 * classes in pairs only (if at all). The "multiplicity condition" of relative
 * GW theory (the sum of the external multiplicities is equal to...) is not
 * checked however.
 *
 */

bool relinv::next() {
	int i,j;
	int m,n;

	// Do one really trivial check first: if we can't have any nodes
	// because no cohomology classes fit on it, exit immediately and
	// signal the end of all splittings.
	if ( node_minclass>xdim ) return 0;

BEGIN:	// If there is only one component, we can go to the degree splitting
	// immediately.
	if (ncpt==1) goto DEGREE;

	// Change node configuration.
	if (node_next()) goto DONE;

AGAIN:	// No further node configuration, change class splittings.
	for (n=ncls-1; n>2; n--) {
		m = 1; // the 0th component is always on its own, so don't
		       // have to check it for splitting within same group
		while (m<ncpt) {
			if (same_next (m,n)) goto ININOD; // -> ok, so init
							  //    nodes
			m += same(m,n-1);
		}
		if (group_next(n)) goto ININOD; // -> ok, so init nodes
	}

	// All class splittings done.

	// If the total genus is 0, we can skip all the genus fiddling and
	// save a lot of time.
	if (totalgenus) {
		// Change loop configuration.
		if (loop_next()) goto ININOD; // -> New loop configuration ok,
					      //    so init nodes. Classes
					      //    need not be initialized.

		// All loop configurations done. Change genus splitting.
		// This is the same as for the class splittings above.
		m = 0;
		while (m<ncpt) {
			if (same_next (m,1)) goto ININOD; // -> ok, so init
							  //    nodes
			m += same(m,0);
		}
		if (group_next(1)) goto ININOD; // -> ok, so init nodes
		// All genus splittings done.
	}

DEGREE:	// Change degree splitting.
	if (deg_next ()) goto ININOD; // New degree split ok, so init nodes.
				      // Note that genus, loops, classes need
				      // not be initialized; they reset
				      // themselves after a complete loop.

	// End of all splittings
	return 0;

ININOD:	// Check if there are classes in the internal component that have
	// too big H power. If there are such classes, move them to component
	// 1 immediately. (We can only get here if we have at least two
	// components.)

	n = 0; // have we changed anything?
	for (i=3; i<ncls; i++) {
		if ( cls[i].Hpow>xdim && value(0,i)) {
			coeff *= factorial (value(0,i));
			cpt_vdim[1] += value(0,i)*cls_vdim[i];
			cpt_tm[1] += value(0,i)*cls[i].mult;
			value(1,i) += value(0,i);
			value(0,i) = 0;
			coeff /= factorial (value(1,i));
			n = 1;
		}
		// Update same value if we have changed anything, which means
		// that components 1 and 2 are definitely different now.
		if (n && same(1,i)>1) {
			same(2,i) = same(1,i)-1;
			same(1,i) = 1;
		}
	}

	// Initialize the nodes.
	if (!node_init()) goto AGAIN;

DONE:	// One last check for too big H powers and wrong sum of multiplicities
	if (!mult()) goto BEGIN;

	#ifdef RELINV_DEBUG
	if (lvl<=RELINV_DEBUG)
		cout << "*relinv* level " << lvl << ": " << *this << endl;
	#endif

	return 1;
}



/*
 * Get component number n. Enough space must have been allocated for the
 * invariant to fit. If n=0, we set the entries iv.pt[i].mult to the
 * multiplicities needed for higher genus relative splitting, i.e. the nodes
 * get minus their external multiplicities, and the internal points get plus
 * their given multiplicities. Still this is always an absolute invariant
 * though, so you MUST call iv.delmult() before passing the invariant to
 * compute() or similar functions.
 *
 */

void relinv::cpt (inv& iv, int n) {
	int i,j;

	iv.xdim = xdim;
	iv.xdeg = xdeg;
	iv.cdeg = cdeg(n);
	iv.cgenus = cgenus(n)-cloop(n);
	iv.points = 0;

	if (n) {
		// Get an external component. First the nodes...
		for (i=0; i<=cloop(n); i++) {
			iv.pt[iv.points].Hpow = node_class(n,i);
			iv.pt[iv.points].psipow = 0;
			iv.pt[iv.points++].mult = node_mult(n,i);
		}
		// ... and then the other classes.
		for (i=3; i<ncls; i++) for (j=0; j<value(n,i); j++)
			iv.pt[iv.points++] = cls[i];
	} else {
		// Get the internal component. First the nodes...
		for (i=1; i<ncpt; i++) for (j=0; j<=cloop(i); j++) {
			if (node_class(i,j)>=0)
				// H power
				iv.pt[iv.points].Hpow = xdim-node_class(i,j);
			else
				// primitive class
				iv.pt[iv.points].Hpow = node_class(i,j);
			iv.pt[iv.points].psipow = 0;
			iv.pt[iv.points++].mult = -node_mult(i,j);
		}
		// ... then the skipped class, if any...
		if (skip)
			iv.pt[iv.points++] = skipclass;
		// ... and then the other classes.
		for (i=3; i<ncls; i++) for (j=0; j<value(0,i); j++)
			iv.pt[iv.points++] = cls[i];
	}
}



// Compute the numerical coefficient of the current splitting.

rational relinv::cof ()
{
	int i,j,k,c;

	// The factors from the class splittings:
	rational result = coeff;

	// The factors from the same components:
	i = 1;
	while (i<ncpt) {
		if (same(i,ncls+1)>1) result /= factorial(same(i,ncls+1));
		i += same(i,ncls+1);
	}

	// The factors from the same nodes:
	if (totalgenus) for (i=1; i<ncpt; i++) {
		j = 1; // nodes in the current node_same block
		for (k=0; k<=cloop(i); k++) {
			if (!(--j)) {
				j = node_same(i,k);
				c = 0;
			}
			if (k>0 && node_mult(i,k-1)!=node_mult(i,k)) c = 0;
			if (++c>1) result /= c;
		}
	}

	// The factors from the diagonal splittings and the multiplicities:
	for (i=1; i<ncpt; i++) for (j=0; j<=cloop(i); j++) {
		if (node_class (i,j)>=0) result /= xdeg;
		if (node_mult(i,j)>1) result *= node_mult(i,j);
	}

	// The factors from the number of primitive classes: this depends
	// on the values of cpt_prim[i].
	if (xdeg>1 && (totalgenus || cpt_prim[0])) {
		j = cpt_prim[1];
		if ( node_class(1,0)<j ) j = node_class(1,0); // the lowest
					// used primitive class index
		if ( prims+j<0 ) return 0; // -> we have used more different
					   // primitive classes than there are
		for (i=cpt_prim[0]; i>j; i--) result *= (prims+i);
		for (i=1; i<ncpt; i++) {
			result /= factorial (cpt_prim[i]-j);
			j = cpt_prim[i];
		}
	}

	return result;
}



// We return 1 if the current splitting satisfies the multiplicity
// condition of relative GW theory, and if there are no classes of too big
// dimension on the internal component.

int relinv::mult ()
{
	int i,j;

	int n = xdeg*cdeg(0);

	// Check H powers of the internal component
	for (i=3; i<ncls; i++) if (cls[i].Hpow>xdim && value(0,i)) return 0;

	// The external multiplicities
	for (i=1; i<ncpt; i++) for (j=0; j<=cloop(i); j++) n += node_mult(i,j);

	// The internal multiplicities
	for (i=3; i<ncls; i++) if (cls[i].mult) n -= value(0,i)*cls[i].mult;
	if (skip) n -= skipclass.mult;

	return !n;
}



ostream& operator<< (ostream& os, relinv& x)
{
	int i,j;

	// Allocate memory for the individual invariants.
	i = x.points;
	for (j=0; j<x.ncpt; j++) i+= x.cloop(j)+1;
	inv iv(i);

	cout << x.cof() << '*';
	for (i=0; i<x.ncpt; i++) {
		x.cpt (iv,i);
		if (!i) iv.delmult();
		cout << iv;
	}

	return os;
}

