/*
 * relsplit.cc - computing relative Gromov-Witten invariants using the
 *               splitting technique
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "relsplit.h"

/*
 * If we include a file setting RELSPLIT_DEBUG to a positive integer, debugging
 * output is printed for all computations in this file of invariants up to
 * this level. This shows how the invariants are computed.
 * 
 */

#include "config.h"



// Debugging macros

#define DEBUG_BEFORE(TEXT)						\
if (lvl<=RELSPLIT_DEBUG)						\
	cout << "*relsplit* level " << lvl				\
	     << ": trying " TEXT << endl;
#define DEBUG_AFTER							\
if (lvl<=RELSPLIT_DEBUG)						\
	cout << "*relsplit* level " << lvl				\
	     << ": result is " << result << endl;

// Compute a relative genus-0 invariant using relative splitting. We
// assume that the last marked point is a relative one. If skip is set to
// a positive number, we skip this many correction terms at the beginning.

rational rs_genus0_rel (inv& iv, int lvl, int skip)
{
	rational result = 0;
	rational term;
	int i;
	int j = iv.points-1;
	int xd = 0;
	int ivdeg = iv.xdeg;
	int ivdim = iv.xdim;

	// Prepare the correction terms.
	iv.pt[j].mult--;
	relinv rv(iv,lvl,j);
	inv iw(iv.points+iv.cdeg);

	// The correction terms.
	do {
		if ( --skip>=0 ) continue;
		if ( !rv.mult() ) continue; // multiplicity condition wrong

		// The internal component.
		rv.cpt (iw,0);
		iw.delmult();
		term = compute (iw,lvl+1);
		if ( is_zero(term) ) continue;

		// The external components. Start from the back, as these are
		// the "easiest" ones. If they turn out to be zero, we can
		// skip the rest.
		for (i=rv.cpts(); i>=1; i--) {
			rv.cpt (iw,i);
			term *= compute (iw,lvl+1);
			if ( is_zero(term) ) break;
		}

		if ( is_zero(term) ) continue;

		#ifdef RELSPLIT_DEBUG
		if (lvl<=RELSPLIT_DEBUG)
			cout << "*relsplit* level " << lvl
			     << ": get term " << -rv.cof()*term << endl;
		#endif

		result -= rv.cof()*term;
		if ( is_undefined(result) ) return result;
	} while ( rv.next() );

	// Decrease multiplicity. If all multiplicities are then gone, change
	// the space to projective space of dimension one bigger.
	if (!iv.pt[j].mult && (!j || !iv.pt[j-1].mult)) {
		xd = iv.xdeg;
		iv.xdeg = 1;
		iv.xdim++;
	}

	// The main H term
	if (iv.pt[j].Hpow>=0)
	if (iv.pt[j].Hpow<ivdim || (iv.pt[j].Hpow<=ivdim &&
				!iv.pt[j].mult)) {
		#ifdef RELSPLIT_DEBUG
		if (lvl<=RELSPLIT_DEBUG)
			cout << "*relsplit* level " << lvl
			     << ": compute main H term" << endl;
		#endif
		iv.pt[j].Hpow++;
		result += (term = ivdeg * compute (iv,lvl+1));
		iv.pt[j].Hpow--;
		#ifdef RELSPLIT_DEBUG
		if (lvl<=RELSPLIT_DEBUG)
			cout << "*relsplit* level " << lvl
			     << ": get term " << term << endl;
		#endif
	}

	// The main psi term
	if (iv.pt[j].mult) {
		#ifdef RELSPLIT_DEBUG
		if (lvl<=RELSPLIT_DEBUG)
			cout << "*relsplit* level " << lvl
			     << ": compute main psi term" << endl;
		#endif
		iv.pt[j].psipow++;
		result += (term = iv.pt[j].mult * compute (iv,lvl+1));
		iv.pt[j].psipow--;
		#ifdef RELSPLIT_DEBUG
		if (lvl<=RELSPLIT_DEBUG)
			cout << "*relsplit* level " << lvl
			     << ": get term " << term << endl;
		#endif
	}

	// Reset the space
	if (xd) {
		iv.xdeg = xd;
		iv.xdim--;
	}
	iv.pt[j].mult++;

	return result;
}



// Compute a genus-0 absolute invariant (of a hypersurface) using relative
// splittings. We need not have a marked point. The trick is to consider
// this as a relative invariant with multiplicity one bigger than allowed.

rational rs_genus0_abs (inv& iv, int lvl)
{
	rational result;
	int f = 0; // do we have to add a marked point? (see below)

	// If there is no marked point, add one using the divisor axiom.
	if (!iv.points) {
		f = 1;
		iv.realloc (1); // allocate memory if necessary
		iv.points = 1;
		iv.pt[0] = pdata(1);
	}

	// Set the multiplicity at the last point to one bigger than allowed.
	iv.pt[iv.points-1].mult = iv.cdeg*iv.xdeg+1;

	// Compute the invariant.
	result = rs_genus0_rel (iv,lvl,1);

	// Reset the invariant.
	iv.pt[iv.points-1].mult = 0;
	if (f) {
		iv.points = 0;
		result /= iv.cdeg; // compensate the divisor axiom
	}

	return result;
}



// Check if we can compute an invariant by relative splitting. Return
// UNDEFINED if this doesn't work. In this case, return the invariant
// unchanged. If the computation succeeds, we store the result in the database.

rational relsplit (inv& iv, int lvl)
{
	rational result;
	int i;

	// Compute relative genus-0 invariants.
	if ( !iv.cgenus && iv.points && iv.pt[iv.points-1].mult) {
		#ifdef RELSPLIT_DEBUG
		DEBUG_BEFORE ("relative splitting for relative g=0 "
				"invariants");
		#endif
		result = rs_genus0_rel (iv,lvl,0);
		#ifdef RELSPLIT_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Compute absolute genus-0 invariants.
	if ( !iv.cgenus && (!iv.points || !iv.pt[iv.points-1].mult)) {
		#ifdef RELSPLIT_DEBUG
		DEBUG_BEFORE ("relative splitting for absolute g=0 "
				"invariants");
		#endif
		result = rs_genus0_abs (iv,lvl);
		#ifdef RELSPLIT_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}
PSI:
	// Sorry, but we can't compute the invariant here. Exit.
	return rational ();

STORE:	// We've successfully computed the invariant. Store it in the database.
	GWdbm.reset_key ();
	GWdbm.reset_value ();
	set_key (GWdbm, iv);
	set_value (GWdbm, result);
	GWdbm.write ();
	return result;
}

