/*
 * trr.cc - implementing the topological recursion relations for any genus
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "trr.h"
#include "split2.h"
#include "linsys.h"

/*
 * If we include a file setting TRR_DEBUG to a positive integer, debugging
 * output is printed for all computations in this file of invariants up to
 * this level. This shows how the invariants are computed.
 * 
 */

#include "config.h"



// Debugging macros

#define DEBUG_BEFORE(TEXT)						\
if (lvl<=TRR_DEBUG)							\
	cout << "*trr* level " << lvl					\
	     << ": trying " TEXT << endl;
#define DEBUG_AFTER							\
if (lvl<=TRR_DEBUG)							\
	cout << "*trr* level " << lvl					\
	     << ": result is " << result << endl;



// Compute a negative genus invariant, using the recursion explained in
// trr.h

rational negative_genus (inv& x, int lvl) {
	rational result;

	inv iv(x), iw(x.points);

	// The psi term
	iv.cgenus++; iv.pt[0].psipow++;
	#ifdef TRR_DEBUG
	if (lvl<=TRR_DEBUG)
		cout << "*trr* level " << lvl
		     << ": compute psi term " << iv << endl;
	#endif
	result = compute (iv, lvl+1);
	#ifdef TRR_DEBUG
	if (lvl<=TRR_DEBUG)
		cout << "*trr* level " << lvl
		     << ": term is " << result << endl;
	#endif
	iv.pt[0].psipow--;

	// The splitting term
	iw.pt[1] = iv.pt[1];
	iv.pt[1] = iv.pt[0];
	iw.cgenus = iv.cgenus;
	iv.cgenus = 0;
	iw.points = 2;
	#ifdef TRR_DEBUG
	if (lvl<=TRR_DEBUG)
		cout << "*trr* level " << lvl
		     << ": compute splitting terms " << endl;
	#endif
	result -= split2_sum (iv,iw,2,2,0,lvl);

	return result;
}



/*
 * Compute an invariant using the topological recursion relation on the last
 * marked point. We assume that we have at least 3g-1 psi classes at this
 * point. This is basically a copy of split2_sum of split2.cc. The main
 * incompatible difference is that the diagonal class has to be the second
 * class in the negative genus invariant, while it is the first one in
 * split2_sum. In the following code, iv is the negative genus invariant, while
 * iw is the genus-g invariant. Another difference compared to split2_sum is
 * the summing over the psi/node sum.
 *
 */

rational trr_reduce (inv& iv, int lvl)
{
	int i,k;

	// Set up the two invariants
	inv **ar;
	assert (ar = new inv*[4]);
	assert (ar[0] = new inv(iv.points+1));
	assert (ar[1] = new inv(iv.points+1));
	ar[0]->xdim = ar[1]->xdim = iv.xdim;
	ar[0]->xdeg = ar[1]->xdeg = iv.xdeg;
	ar[1]->cgenus = iv.cgenus;
	ar[0]->points = iv.points+1;
	ar[1]->points = 1;
	ar[0]->pt[0] = iv.pt[iv.points-1];
	ar[0]->pt[0].psipow += 1-3*iv.cgenus;
	ar[0]->pt[1].psipow = 0;
	ar[0]->pt[1].mult = ar[1]->pt[0].mult = 0;
	for (i=iv.points-2; i>=0; i--) ar[0]->pt[i+2] = iv.pt[i];

	// Copy the invariants as needed by next_pair
	assert (ar[2] = new inv(*(ar[0])));
	assert (ar[3] = new inv(*(ar[1])));

	rational result = 0;
	rational coeff = rational(1)/iv.xdeg; // factor from diagonal class
	rational tmpv,tmpw;
	int dw;
	int d = iv.cdeg ; // the total degree
	int d1,d2; // the valid degree bounds in the current term for iw

	dw = (iv.xdim-3)*(1-iv.cgenus);

	// Now do the loop over all class splittings...
	do {
		// ... over the psi/node splitting...
		for (k=3*iv.cgenus-2; k>=0; k--) {
			ar[0]->cgenus = k+2-3*iv.cgenus;
			ar[1]->pt[0].psipow = k;
			// ... and over all degree splittings.
			// Compute the bounds for the degrees
			if (iv.xdim+2-iv.xdeg>0) {
				// the iv.xdim+1-iv.xdeg in the numerator is
				// to round up
				d1 = idiv (iv.xdim-iv.xdeg-dw+k,
						iv.xdim+2-iv.xdeg);
				if (d1<0) d1=0;
				d2 = idiv (iv.xdim-1-dw+k,iv.xdim+2-iv.xdeg);
				if (d2>d) d2=d;
			} else if (iv.xdim+2-iv.xdeg<0) {
				// the iv.xdeg-iv.xdim-3 in the numerator is
				// to round up
				d1 = idiv (dw-k-iv.xdim+iv.xdeg-iv.xdim-2,
					iv.xdeg-iv.xdim-2);
				if (d1<0) d1=0;
				d2 = idiv (1+dw-k,iv.xdeg-iv.xdim-2);
				if (d2>d) d2=d;
			} else {
				d1 = 0; d2 = d;
			}
			for (ar[1]->cdeg=d1; ar[1]->cdeg<=d2; ar[1]->cdeg++) {
				ar[0]->cdeg = d-ar[1]->cdeg;

				// fill in the diagonal class
				ar[0]->pt[1].Hpow = iv.xdim -
					(ar[1]->pt[0].Hpow =
						ar[1]->cdeg*(iv.xdim+2-iv.xdeg)
						+dw-k+1);

				#ifdef TRR_DEBUG
				if (ar[0]->pt[1].Hpow<0 || ar[1]->pt[0].Hpow<0)
					cout << "*trr* level " << lvl
					     << ": this shouldn't happen"
					     << endl;
				#endif

				// Compute the two invariants. Take the
				// rational (negative genus) one first.
				#ifdef TRR_DEBUG
				if (lvl<=TRR_DEBUG)
					cout << "*trr* level " << lvl
					     << ": need term " << coeff << "*"
					     << *ar[0] << "*" << *ar[1]
					     << endl;
				#endif
				tmpv = compute (*ar[0], lvl+1);
				if ( is_zero(tmpv) ) continue;
				tmpw = compute (*ar[1], lvl+1);
				if ( is_zero(tmpw) ) continue;
				result += coeff*tmpv*tmpw;
				#ifdef TRR_DEBUG
				if (lvl<=TRR_DEBUG)
					cout << "*trr* level " << lvl
					     << ": term is "
					     << coeff*tmpv*tmpw << endl;
				#endif
			}
			if ( is_undefined (result) ) break;
		}
	} while ( next_pair (ar,2,1,dw,coeff) );

	delete ar[0];
	delete ar[1];
	delete ar[2];
	delete ar[3];
	delete ar;

	return result;
}



/*
 * Same as trr_reduce, but the genus-g invariants with full marked points are
 * only checked with 'simplify' and not calculated recursively with 'compute'.
 * If the invariant cannot be calculated this way, it is written as a variable
 * in a linear equation. This is needed for the Virasoro conditions.
 *
 * Moreover, we use the topological recursion relation on the first marked
 * point, instead of on the last.
 *
 * The function returns the constant part of the linear equation. The variable
 * parts are written to the linear system Q.
 *
 * If the invariant cannot be simplified using the topological recursion
 * relations, it is returned as-is (i.e. as a variable).
 *
 * The rational coefficient cof is taken as an overall factor for the whole
 * expression (again this is needed for the Virasoro conditions).
 *
 */

rational trr_variable (linsys& Q, inv& iv, rational& cof, int lvl)
{
	rational result;
	int i,k;
	inv iw;

	// Check if the invariant is known already. Be careful to NOT sort
	// the invariant.
	iw = iv;
	result = simplify (iw,lvl+1);
	if ( !is_undefined(result) ) {
		#ifdef TRR_DEBUG
		if (lvl<=TRR_DEBUG)
			cout << "*trr* level " << lvl
			     << ": term is " << cof*result
			     << endl;
		#endif
		return cof*result;
	}

	// Check if we can reduce the invariant.
	if ( iv.pt[0].psipow<3*iv.cgenus-1 ) {
		#ifdef TRR_DEBUG
		if (lvl<=TRR_DEBUG)
			cout << "*trr* level " << lvl
			     << ": term is " << cof
			     << "x[" << iv.cdeg*(iv.xdim+1)+iv.pt[0].Hpow
			     << "]" << endl;
		#endif
		if ( Q.add_var_term (iv.cdeg*(iv.xdim+1)+iv.pt[0].Hpow,-cof)
				<0 ) {
			cout << "*trr* level " << lvl
			     << ": too many variables!" << endl;
			return rational ();
		}
		return rational (0);
	}

	#ifdef TRR_DEBUG
	if (lvl<=TRR_DEBUG)
		cout << "*trr* level " << lvl
		     << ": reducing invariant " << iv << endl;
	#endif

	// Set up the two invariants
	inv **ar;
	assert (ar = new inv*[4]);
	assert (ar[0] = new inv(iv.points+1));
	assert (ar[1] = new inv(iv.points+1));
	ar[0]->xdim = ar[1]->xdim = iv.xdim;
	ar[0]->xdeg = ar[1]->xdeg = iv.xdeg;
	ar[1]->cgenus = iv.cgenus;
	ar[0]->points = iv.points+1;
	ar[1]->points = 1;
	ar[0]->pt[0] = iv.pt[0];
	ar[0]->pt[0].psipow += 1-3*iv.cgenus;
	ar[0]->pt[1].psipow = 0;
	ar[0]->pt[1].mult = ar[1]->pt[0].mult = 0;
	for (i=iv.points-1; i>0; i--) ar[0]->pt[i+1] = iv.pt[i];

	// Copy the invariants as needed by next_pair
	assert (ar[2] = new inv(*(ar[0])));
	assert (ar[3] = new inv(*(ar[1])));

	result = 0;
	rational coeff = rational(1)/iv.xdeg; // factor from diagonal class
	rational tmpv,tmpw;
	int dw;
	int d = iv.cdeg ; // the total degree
	int d1,d2; // the valid degree bounds in the current term for iw

	dw = (iv.xdim-3)*(1-iv.cgenus);

	// Now do the loop over all class splittings...
	do {
	    // ... over the psi/node splitting...
	    for (k=3*iv.cgenus-2; k>=0; k--) {
		ar[0]->cgenus = k+2-3*iv.cgenus;
		ar[1]->pt[0].psipow = k;
		// ... and over all degree splittings.
		// Compute the bounds for the degrees. The
		// iv.xdim+1-iv.xdeg in the numerator is to round up
		d1 = idiv (iv.xdim-iv.xdeg-dw+k,iv.xdim+2-iv.xdeg);
		if (d1<0) d1=0;
		d2 = idiv (iv.xdim-1-dw+k,iv.xdim+2-iv.xdeg);
		if (d2>d) d2=d;

		for (ar[1]->cdeg=d1; ar[1]->cdeg<=d2; ar[1]->cdeg++) {
		    ar[0]->cdeg = d-ar[1]->cdeg;

		    // fill in the diagonal class
		    ar[0]->pt[1].Hpow = iv.xdim -
			(ar[1]->pt[0].Hpow =
				ar[1]->cdeg*(iv.xdim+2-iv.xdeg)+dw-k+1);

		    #ifdef TRR_DEBUG
		    if (ar[0]->pt[1].Hpow<0 || ar[1]->pt[0].Hpow<0)
			cout << "*trr* level " << lvl
			     << ": this shouldn't happen" << endl;
		    #endif

		    // Compute the two invariants. Take the
		    // rational (negative genus) one first.
		    #ifdef TRR_DEBUG
		    if (lvl<=TRR_DEBUG)
		 	cout << "*trr* level " << lvl
			     << ": need term " << coeff << "*"
			     << *ar[0] << "*" << *ar[1]
			     << endl;
		    #endif

		    tmpv = compute (*ar[0], lvl+1);
		    if ( is_zero(tmpv) ) continue;
		    if (ar[0]->points!=2) {
			tmpw = compute (*ar[1], lvl+1);
			if ( is_zero(tmpw) ) continue;
		    } else {
			// Be careful NOT to sort the invariant
			iw = *ar[1];
			tmpw = simplify (iw, lvl+1);
			if ( is_zero(tmpw) ) continue;
			if ( is_undefined(tmpw) ) {
			    #ifdef TRR_DEBUG
			    if (lvl<=TRR_DEBUG)
				cout << "*trr* level " << lvl
				     << ": term is " << cof*coeff*tmpv
				     << "x["
				     << ar[1]->cdeg*(iv.xdim+1)
				     	+ar[1]->pt[0].Hpow 
				     << "]" << endl;
			    #endif
			    if ( Q.add_var_term (ar[1]->cdeg*(iv.xdim+1)
				     +ar[1]->pt[0].Hpow,
				     -cof*coeff*tmpv) <0 ) {
				    cout << "*trr* level " << lvl
				         << ": too many variables!"
				         << endl;
				return rational ();
			    }
			    continue;
			}
		    }
		    result += coeff*tmpv*tmpw;

		    #ifdef TRR_DEBUG
		    if (lvl<=TRR_DEBUG)
		    cout << "*trr* level " << lvl
		         << ": term is "
		         << cof*coeff*tmpv*tmpw << endl;
		    #endif
		}
		if ( is_undefined (result) ) break;
	    }
	} while ( next_pair (ar,2,1,dw,coeff) );

	delete ar[0];
	delete ar[1];
	delete ar[2];
	delete ar[3];
	delete ar;

	return cof*result;
}



// Check if we can compute an invariant using the topological recursion
// relations. Return UNDEFINED if this doesn't work. In this case, return the
// invariant unchanged. If the computation succeeds, we store the result in
// the database.

rational trr_genusg (inv& iv, int lvl)
{
	rational result;

	// Relative invariants are not supported here.
	if ( iv.points && iv.pt[iv.points-1].mult ) return rational ();

	// Currently, invariants with primitive classes are not supported
	// either.
	if ( (iv.xdeg>1) && iv.reduce() ) return rational ();

	// Compute the negative genus invariants.
	if ( iv.cgenus<0 ) {
		#ifdef TRR_DEBUG
		DEBUG_BEFORE ("negative genus recursion");
		#endif
		result = negative_genus (iv,lvl);
		#ifdef TRR_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// If there are enough psi classes at one point, use the topological
	// recursion relations.
	if ( (iv.cgenus>0) && iv.points &&
			(iv.pt[iv.points-1].psipow >= 3*iv.cgenus-1) ) {
		#ifdef TRR_DEBUG
		DEBUG_BEFORE ("general topological recursion relation");
		#endif
		result = to_rational (trr_reduce (iv,lvl));
		#ifdef TRR_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Sorry, but we can't compute the invariant here. Exit.
	return rational ();

STORE:	// We've successfully computed the invariant. Store it in the database.
	GWdbm.reset_key ();
	GWdbm.reset_value ();
	set_key (GWdbm, iv);
	set_value (GWdbm, result);
	GWdbm.write ();
	return result;
}

