/*
 * virasoro.cc - computing GW invariants using the Virasoro conditions
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "virasoro.h"
#include "symfunc.h"
#include "split2.h"
#include "trr.h"
#include "linsys.h"

/*
 * If we include a file setting VIRASORO_DEBUG to a positive integer,
 * debugging output is printed for all computations in this file of
 * invariants up to this level. This shows how the invariants are computed.
 * Setting VIRASORO_LINSYS_DEBUG to a positive integer has a similar effect,
 * but prints only the final linear equation for every L[k] checked, and
 * how the invariants are determined from this.
 * 
 * We also have to set VIRASORO_TRY to a non-negative integer. This specifies
 * how often we can hit a linearly dependent Virasoro condition before we give
 * up and return UNDEFINED.
 *
 * We must also set VIRASORO_CHECK to a non-negative integer. This specifies
 * how many additional Virasoro conditions we'll check after the system is
 * solvable. Set it to 0 for maximum performance, or to a positive value to
 * check the algorithm.
 * 
 */

#include "config.h"



// Debugging macros

#define DEBUG_BEFORE(TEXT)						\
if (lvl<=VIRASORO_LINSYS_DEBUG)						\
	cout << "*virasoro* level " << lvl				\
	     << ": trying " TEXT << endl;
#define DEBUG_AFTER							\
if (lvl<=VIRASORO_LINSYS_DEBUG)						\
	cout << "*virasoro* level " << lvl				\
	     << ": result is " << result << endl;



// The following functions compute the various terms in the Virasoro condition.
// iv,d corresponds to the t,q-power in which we look at the conditions.
// Variables are stored in the linear system Q. They can only occur in the A
// and D terms.

rational virasoro_A (linsys& Q, inv& iv, int lvl, int k, int d)
{
	int i;
	rational R = 1;
	rational term;
	rational result = 0;

	inv iw(iv.points+1);
	iw.points = iv.points+1;
	for (i=iv.points-1; i>=0; i--) iw.pt[i+1] = iv.pt[i];
	iw.pt[0].mult = 0;
	iw.xdim = iv.xdim;
	iw.xdeg = iv.xdeg;
	iw.cgenus = iv.cgenus;
	iw.cdeg = d;

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": computing A terms" << endl;
	#endif

	for (i=0; i<=k+1 && i<=iv.xdim; i++) {
		iw.pt[0].Hpow = i;
		iw.pt[0].psipow = k+1-i;
		term = -R*symfunc(rational(3-iv.xdim)/2,k,i);
		R *= (iv.xdim+2-iv.xdeg);
		#ifdef VIRASORO_DEBUG
		if (lvl<=VIRASORO_DEBUG)
			cout << "*virasoro* level " << lvl
 			     << ": need term " << term
			     << "*" << iw << endl;
		#endif
		if ( is_zero(term) ) continue;
		term = trr_variable (Q, iw, term, lvl);
		if ( is_undefined (term) ) return rational ();
		result += term;
	}

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": total constant A terms are " << result << endl;
	#endif

	return result;
}



rational virasoro_B (inv& iv, int lvl, int k, int d)
{
	int a,dd,i,j,m;
	int cnt = 1;
	rational R;
	rational term;
	rational result = 0;

	inv iw(iv);
	iw.cdeg = d;

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": computing B terms" << endl;
	#endif

	for (j=0; j<iw.points; j++) {
		// If two classes are equal, compute the terms only once
		if ( j<iw.points-1 && iw.pt[j]==iw.pt[j+1] ) {
			cnt++;
			continue;
		}

		a = iw.pt[j].Hpow;
		m = iw.pt[j].psipow;
		R = 1;
		for (i=0; i<=k+1 && a+i<=iw.xdim; i++) {
			if (i) R *= (iw.xdim+2-iw.xdeg);
			iw.pt[j].Hpow = a+i;
			if ( (iw.pt[j].psipow = k+m-i) >= 0 ) {
				term = R*cnt*symfunc
					(rational(-iw.xdim+1)/2+a+m,k,i);
				#ifdef VIRASORO_DEBUG
				if (lvl<=VIRASORO_DEBUG)
					cout << "*virasoro* level " << lvl
		 			     << ": need term " << term
					     << "*" << iw << endl;
				#endif
				if ( is_zero(term) ) continue;
				term *= compute (iw,lvl+1);
				if ( is_undefined (term) ) return rational ();
				#ifdef VIRASORO_DEBUG
				if (lvl<=VIRASORO_DEBUG)
					cout << "*virasoro* level " << lvl
		 			     << ": term is " << term
					     << endl;
				#endif
				result += term;
			}
		}
		iw.pt[j].Hpow = a;
		iw.pt[j].psipow = m;
		cnt = 1;
	}

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": total B terms are " << result << endl;
	#endif

	return result;
}



// In the C terms we use the symmetry between the two added points to reduce
// computation time.

rational virasoro_C (inv& iv, int lvl, int k, int d)
{
	int a,i,m;
	rational R = 1;
	rational term;
	rational result = 0;
	inv iw(iv.points+2);
	iw.points = iv.points+2;
	for (i=iv.points-1; i>=0 ;i--) iw.pt[i+2] = iv.pt[i];
	iw.pt[0].mult = iw.pt[1].mult = 0;
	iw.xdim = iv.xdim;
	iw.xdeg = iv.xdeg;
	iw.cgenus = iv.cgenus-1;
	iw.cdeg = d;

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": computing C terms" << endl;
	#endif

	for (i=0; i<=k+1 && i<=iv.xdim; i++) {
		for (m=idiv(i-k-1,2); m<=-1; m++) {
			for (a=i; a<=iv.xdim; a++) {
				iw.pt[0].Hpow = a;
				iw.pt[0].psipow = -m-1;
				iw.pt[1].Hpow = iv.xdim+i-a;
				iw.pt[1].psipow = m+k-i;
				if (iw.pt[0]>iw.pt[1]) continue;
				term = R*symfunc
					(rational(iw.xdim+1)/2-a+m,k,i);
				if (m%2) term = -term;
				if (iw.pt[0]==iw.pt[1]) term /= 2;
				#ifdef VIRASORO_DEBUG
				if (lvl<=VIRASORO_DEBUG)
					cout << "*virasoro* level " << lvl
		 			     << ": need term " << term
					     << "*" << iw << endl;
				#endif
				if ( is_zero(term) ) continue;
				term *= compute (iw,lvl+1);
				if ( is_undefined (term) ) return rational ();
				#ifdef VIRASORO_DEBUG
				if (lvl<=VIRASORO_DEBUG)
					cout << "*virasoro* level " << lvl
		 			     << ": term is " << term
					     << endl;
				#endif
				result += term;
			}
		}
		R *= (iv.xdim+2-iv.xdeg);
	}

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": total C terms are " << result << endl;
	#endif

	return result;
}



// In the D terms we use the h <-> g-h symmetry to reduce computation time.

rational virasoro_D (linsys& Q, inv& iv, int lvl, int k, int d)
{
	int i,m;
	int skip = 1;
	rational R;
	rational fact;

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": computing D terms" << endl;
	#endif

	// Space for the 4 invariants needed by next_pair
	inv **ar;
	assert (ar = new inv*[4]);
	assert (ar[0] = new inv(iv.points+1));
	assert (ar[1] = new inv(iv.points+1));
	ar[0]->xdim = ar[1]->xdim = iv.xdim;
	ar[0]->xdeg = ar[1]->xdeg = iv.xdeg;
	ar[0]->points = iv.points+1;
	ar[1]->points = 1;
	ar[0]->pt[0].mult = ar[1]->pt[0].mult = 0;
	for (i=iv.points-1; i>=0; i--) ar[0]->pt[i+1] = iv.pt[i];

	// Copy the invariants as needed by next_pair
	assert (ar[2] = new inv(*(ar[0])));
	assert (ar[3] = new inv(*(ar[1])));

	rational result = 0;
	rational coeff = rational(1)/iv.xdeg; // factor from diagonal class
	rational tmpv,tmpw;
	int ddw,dw;
	int d1,d2; // the valid degree bounds in the current term for iw
	int g;

	ddw = iv.xdim-3;

	// Now do the loop over all class splittings...
	do {
	    // ... over the genus splitting...
	    for (g=0; g<=iv.cgenus/2; g++) {
		ar[0]->cgenus = iv.cgenus-g;
		ar[1]->cgenus = g;
		// ... over the i loop...
		R = 1;
		for (i=0; i<=k+1 && i<=iv.xdim; i++) {
		    // ... over the m loop...
		    for (m=i-k; m<=-1; m++) {
			ar[0]->pt[0].psipow = -m-1;
			ar[1]->pt[0].psipow = m+k-i;
			// ... and the loop over all degree splittings.
			// Compute the bounds for the degrees.
			dw = ddw-(iv.xdim-3)*g-m-k+i;
			// the iv.xdim+1-iv.xdeg in the numerator is to round
			// up
			d1 = idiv (iv.xdim-iv.xdeg-dw+i,iv.xdim+2-iv.xdeg);
			if (d1<0) d1=0;
			d2 = idiv (iv.xdim-1-dw,iv.xdim+2-iv.xdeg);
			if (d2>d) d2=d;

			for (ar[1]->cdeg=d1; ar[1]->cdeg<=d2; ar[1]->cdeg++) {
			    ar[0]->cdeg = d-ar[1]->cdeg;

			    // fill in the diagonal class
			    ar[0]->pt[0].Hpow = iv.xdim+i -
					(ar[1]->pt[0].Hpow =
			    	ar[1]->cdeg*(iv.xdim+2-iv.xdeg)+dw+1);

			    #ifdef VIRASORO_DEBUG
			    if (ar[0]->pt[0].Hpow<0 || ar[1]->pt[0].Hpow<0)
				cout << "*virasoro* level " << lvl
				     << ": this shouldn't happen"
				    << endl;
			    #endif

			    fact = R*symfunc (rational(iv.xdim+1)/2
					-ar[0]->pt[0].Hpow+m,k,i);
			    if (g+g==iv.cgenus) fact /= 2;
			    if (m%2) fact = -fact;

			    // Compute the two invariants.
			    #ifdef VIRASORO_DEBUG
			    if (lvl<=VIRASORO_DEBUG)
				cout << "*virasoro* level "
				     << lvl
				     << ": need term "
				     << fact << "*"
				     << coeff << "*"
				     << *ar[0] << "*"
				     << *ar[1] << endl;
			    #endif

			    if ( is_zero(fact) ) continue;
			    tmpw = compute (*ar[1], lvl+1);
			    if ( is_zero(tmpw) ) continue;
			    tmpw *= fact*coeff;

			    if (skip) {
				// full genus and all marked points
				tmpv = trr_variable (Q, *ar[0], tmpw, lvl);
				if ( is_undefined(tmpv) ) return rational ();
				result += tmpv;
				continue;
			    }

			    tmpv = compute (*ar[0], lvl+1);
			    if ( is_zero(tmpv) ) continue;

			    result += tmpv*tmpw;

			    #ifdef VIRASORO_DEBUG
			    if (lvl<=VIRASORO_DEBUG)
				cout << "*virasoro* level "
				     << lvl << ": term is "
				     << tmpv*tmpw
				     << endl;
			    #endif
			}
		    }
		    R *= (iv.xdim+2-iv.xdeg);
		}
		skip = 0;
	    }
	    if ( is_undefined (result) ) break;
	} while ( next_pair (ar,1,1,ddw,coeff) );

	delete ar[0];
	delete ar[1];
	delete ar[2];
	delete ar[3];
	delete ar;

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": total constant D terms are " << result << endl;
	#endif

	return result;
}



/*
 * Evaluate the L_k Virasoro condition and store the resulting equation in the
 * linear system. The invariant iv is the t power at which we evaluate the
 * Virasoro condition, so the variables in the linear system correspond to
 * genus-g invariants with the classes in iv plus one more class. If this
 * one more class is H^i psi^j and the degree is d, the invariant corresponds
 * to the variable with the name (iv.xdim+1)*d+i.
 *
 */

enum linsys_t virasoro_k (linsys& Q, inv& iv, int lvl, int k, int d)
{
	rational x;
	enum linsys_t t;

	#ifdef VIRASORO_DEBUG
	x = virasoro_A (Q, iv, lvl, k, d);
	if ( is_undefined(x) ) return UNDETERMINED;
	Q.add_const_term (x);
	if (lvl<=VIRASORO_DEBUG) {
		cout << "*virasoro* level " << lvl
		     << ": current equation is ";
		Q.write_current_eqn();
	}
	x = virasoro_B (iv, lvl, k, d);
	if ( is_undefined(x) ) return UNDETERMINED;
	Q.add_const_term (x);
	if (lvl<=VIRASORO_DEBUG) {
		cout << "*virasoro* level " << lvl
		     << ": current equation is ";
		Q.write_current_eqn();
	}
	x = virasoro_C (iv, lvl, k, d);
	if ( is_undefined(x) ) return UNDETERMINED;
	Q.add_const_term (x);
	if (lvl<=VIRASORO_DEBUG) {
		cout << "*virasoro* level " << lvl
		     << ": current equation is ";
		Q.write_current_eqn();
	}
	x = virasoro_D (Q, iv, lvl, k, d);
	if ( is_undefined(x) ) return UNDETERMINED;
	Q.add_const_term (x);
	if (lvl<=VIRASORO_DEBUG) {
		cout << "*virasoro* level " << lvl
		     << ": current equation is ";
		Q.write_current_eqn();
	}
	#else
	x = virasoro_A (Q, iv, lvl, k, d) +
	    virasoro_B (iv, lvl, k, d) +
	    virasoro_C (iv, lvl, k, d) +
	    virasoro_D (Q, iv, lvl, k, d);
	if ( is_undefined(x) ) return UNDETERMINED;
	Q.add_const_term (x);
	#endif

	#ifdef VIRASORO_LINSYS_DEBUG
	if (lvl<=VIRASORO_LINSYS_DEBUG) {
		cout << "*virasoro* level " << lvl
		     << ": got equation ";
		Q.write_current_eqn();
	}
	#endif

	t = Q.send_row ();

	#ifdef VIRASORO_DEBUG
	if (lvl<=VIRASORO_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": " << Q << endl;
	#endif

	return t;
}



// Check if we can compute an invariant using the Virasoro conditions.
// Return UNDEFINED if this doesn't work. In this case, return the
// invariant unchanged. If the computation succeeds, we store the result in
// the database.

rational virasoro (inv& iv, int lvl)
{
	int mx,d,i,j,k;
	rational result;
	enum linsys_t t;
	inv iw;

	// We must have a marked point.
	if ( !iv.points ) return rational ();

	// Relative invariants are not supported here.
	if ( iv.pt[iv.points-1].mult ) return rational ();

	// Currently, invariants with primitive classes are not supported
	// either.
	if ( (iv.xdeg>1) && iv.reduce() ) return rational ();

	// The genus needs to be positive to use our algorithm.
	if ( iv.cgenus<=0 ) return rational ();

	// Currently this works in P^n only.
	if ( iv.xdeg!=1 ) return rational ();

	// The psi power at the last marked point must be smaller than 3g-1.
	if ( iv.pt[iv.points-1].psipow>=3*iv.cgenus-1 ) return rational ();

	// Try a Virasoro computation.
	#ifdef VIRASORO_LINSYS_DEBUG
	if (lvl<=VIRASORO_LINSYS_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": trying Virasoro conditions for " << iv << endl;
	#endif

	iw = iv;

	/*
	 * Set up a linear system. First compute the maximum number of
	 * variables that we may need. For this we need the minimum and
	 * maximum possible degree. We get this by remarking that we can have
	 * at most H^(dim) * psi^(3g-2) on the last marked point (and at least
	 * the fundamental class), so we can get at most
	 *   dim+3g-2-(H power of last point)-(psi power of last point)
	 * more conditions, and at least
	 *   (H power of last point)+(psi power of last point)
	 * fewer conditions. In other words, the possible range in terms of
	 * (complex) dimensions is dim+3g-1. To see what this corresponds to in
	 * terms of degrees, we have to divide this number by xdim+2-xdeg,
	 * but make sure that we round up.
	 *
	 */
	mx = (iw.xdim+1) *
		( (3*iw.cgenus+2*iw.xdim-iw.xdeg) / (iw.xdim+2-iw.xdeg) );
	linsys Q(mx);
	Q.add_var ((iw.xdim+1)*iw.cdeg+iw.pt[iw.points-1].Hpow);
	#ifdef VIRASORO_LINSYS_DEBUG
	if (lvl<=VIRASORO_LINSYS_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": reserving space for " << mx << " variables" << endl;
	#endif

	/*
	 * Compute the smallest k for which we can check the Virasoro
	 * condition. For this note that the Virasoro condition L_k is
	 * of codimension k, i.e. it adds k+1 conditions at the last marked
	 * point. So (H power of last point)+(psi power of last point)-1 is
	 * certainly an allowed value for k. The possible steps for k are
	 * xdim+2-xdeg, so we can reduce modulo this number. But note that we
	 * want k at least be equal to 1, so that we don't have to worry about
	 * the (E) terms. The number d is the corresponding degree.
	 *
	 */
	k = iw.pt[iw.points-1].Hpow+iw.pt[iw.points-1].psipow-1;
	d = iw.cdeg;
	d++; k += iw.xdim+2-iw.xdeg;
	d -= k/(iw.xdim+2-iw.xdeg);
	k %= (iw.xdim+2-iw.xdeg);
	if (!k) {d++; k += iw.xdim+2-iw.xdeg;}
	if (d<0) { k -= (iw.xdim+2-iw.xdeg)*d; d=0; }

	// Now check the Virasoro conditions in turn.
	j = 0;
	iw.points--;
	for (i=1; i<=mx+VIRASORO_TRY; i++) {
		#ifdef VIRASORO_LINSYS_DEBUG
		if (lvl<=VIRASORO_LINSYS_DEBUG)
			cout << "*virasoro* level " << lvl
			     << ": computation number " << i
			     << ", checking L[" << k
			     << "] in degree " << d
			     << ", genus " << iw.cgenus << endl;
		#endif

		t = virasoro_k (Q, iw, lvl, k, d);

		if ( t==CONTRADICTION ) {
			cout << "*virasoro* level " << lvl
			     << ": got a contradiction" << endl;
			return rational ();
		}
		if ( t==SOLVABLE ) {
			#ifdef VIRASORO_LINSYS_DEBUG
			if (lvl<=VIRASORO_LINSYS_DEBUG) {
				cout << "*virasoro* level " << lvl
				     << ": " << Q
				     << "*virasoro* level " << lvl
				     << ": the system is solvable now";
				if (j>0) cout << ", " << j << " check(s) OK";
				cout << endl;
			}
			#endif
			j++;
		}
		#ifdef VIRASORO_LINSYS_DEBUG
		if ( (lvl<=VIRASORO_LINSYS_DEBUG) && (t==UNDETERMINED) )
			cout << "*virasoro* level " << lvl
			     << ": " << Q
			     << "*virasoro* level " << lvl
			     << ": the system is not solvable yet" << endl;
		#endif
		if ( j>VIRASORO_CHECK ) break;
		d++; k += iw.xdim+2-iw.xdeg;
	}

	#ifdef VIRASORO_LINSYS_DEBUG
	if (lvl<=VIRASORO_LINSYS_DEBUG)
		cout << "*virasoro* level " << lvl
		     << ": determinant of the system is " << Q.det() << endl;
	#endif

	if ( t==SOLVABLE ) {
		// storing all invariants
		for (i=0; i<Q.numvars(); i++) {
			iw = iv;
			j = (Q.varname(i))/(iw.xdim+1);
			k = (Q.varname(i))%(iw.xdim+1);
			iw.pt[iw.points-1].psipow += iw.pt[iw.points-1].Hpow-k
					+ (iw.xdim+2-iw.xdeg)*(j-iw.cdeg);
			iw.cdeg = j;
			iw.pt[iw.points-1].Hpow = k;
			result = Q.value_by_number (i);
			iw.sort ();

			#ifdef VIRASORO_LINSYS_DEBUG
			if (lvl<=VIRASORO_LINSYS_DEBUG)
				cout << "*virasoro* level " << lvl
				     << ": setting " << iw
				     << " to " << result << endl;
			#endif

			GWdbm.reset_key ();
			GWdbm.reset_value ();
			set_key (GWdbm, iw);
			set_value (GWdbm, result);
			GWdbm.write ();
		}

		result = Q.value_by_name
			((iv.xdim+1)*iv.cdeg+iv.pt[iv.points-1].Hpow);
	} else {
		result = rational ();
		#ifdef VIRASORO_LINSYS_DEBUG
		if (lvl<=VIRASORO_LINSYS_DEBUG)
			cout << "*virasoro* level " << lvl
			     << ": not enough conditions found, giving up"
			     << endl;
		#endif
	}

	#ifdef VIRASORO_LINSYS_DEBUG
	DEBUG_AFTER;
	#endif

	return result;
}

