/*
 * monomial.cc - a class for monomial arithmetic
 *
 */

#include <assert.h>

#include "monomial.h"

/*
 * If we include a file defining MONOMIAL_DEBUG, debugging output is printed
 * whenever the reference counter of a monomial is raised or dropped. The
 * last message printed at program exit should be that the local and global
 * reference counters are both 0.
 *
 */

#include "config.h"



void monomial_data::monomial_debug (int n, int d)
{
	#ifdef MONOMIAL_DEBUG

	static int totalrefcnt = 0;

	totalrefcnt += d;

	cout << "*monomial* local ref count " << n;
	cout << ", global ref count " << totalrefcnt << endl;

	assert (n>=0 && totalrefcnt>=0);

	#endif
}



// Multiplication of monomials 

monomial operator* (const monomial& x, const monomial& y)
{
	monomial result;

	if (!x.ptr || !y.ptr) return result;

	short ix = 0, iy = 0, ir = 0;
	short vx = x.ptr->vars;
	short vy = y.ptr->vars;
	
	/*
	 * Allocate enough space for the variables of both monomials. This
	 * may be too much, but that doesn't matter as long as we set 'vars'
	 * of the result correctly.
	 *
	 */
	assert (result.ptr = new monomial_data (vx+vy));

	while (ix<vx || iy<vy) {
		if (ix>=vx || (iy<vy && y.ptr->vr(iy)<x.ptr->vr(ix))) {
			result.ptr->vr(ir) = y.ptr->vr(iy);
			result.ptr->ex(ir++) = y.ptr->ex(iy++);
			continue;
		};
		if (iy>=vy || (ix<vx && x.ptr->vr(ix)<y.ptr->vr(iy))) {
			result.ptr->vr(ir) = x.ptr->vr(ix);
			result.ptr->ex(ir++) = x.ptr->ex(ix++);
			continue;
		};
		result.ptr->vr(ir) = x.ptr->vr(ix);
		result.ptr->ex(ir++) = x.ptr->ex(ix++)+y.ptr->ex(iy++);
	}

	result.ptr->vars = ir;

	return result;
}

monomial operator* (const monomial& x, const var& v)
	{return x*monomial(v);};
monomial operator* (const var& v, const monomial& x)
	{return monomial(v)*x;};
monomial operator* (const var& v, const var& w)
	{return monomial(v)*monomial(w);};



// Division of monomials

monomial operator/ (const monomial& x, const monomial& y)
{
	monomial result;

	if (!x.ptr || !y.ptr) return result;

	short i, ix = 0, iy = 0, ir = 0;
	short vx = x.ptr->vars;
	short vy = y.ptr->vars;
	
	/*
	 * Allocate enough space for the variables of both monomials. This
	 * may be too much, but that doesn't matter as long as we set 'vars'
	 * of the result correctly.
	 *
	 */
	assert (result.ptr = new monomial_data (vx+vy));

	while (ix<vx || iy<vy) {
		if (ix>=vx || (iy<vy && y.ptr->vr(iy)<x.ptr->vr(ix)))
			return monomial (); // undefined
		if (iy>=vy || (ix<vx && x.ptr->vr(ix)<y.ptr->vr(iy))) {
			result.ptr->vr(ir) = x.ptr->vr(ix);
			result.ptr->ex(ir++) = x.ptr->ex(ix++);
			continue;
		};
		result.ptr->vr(ir) = x.ptr->vr(ix);
		i = result.ptr->ex(ir) = x.ptr->ex(ix++)-y.ptr->ex(iy++);
		if (i<0) return monomial ();
		if (i>0) ir++;
	}

	result.ptr->vars = ir;

	return result;
}

monomial operator/ (const monomial& x, const var& v)
	{return x/monomial(v);};
monomial operator/ (const var& v, const monomial& x)
	{return monomial(v)/x;};
monomial operator/ (const var& v, const var& w)
	{return monomial(v)/monomial(w);};



// Exponentiation of monomials

monomial pow (const monomial& x, const int& n)
{
	monomial result;

	if (!x.ptr || n<0) return result;

	if (!n) return monomial(1);

	short i;
	short vx = x.ptr->vars;
	
	assert (result.ptr = new monomial_data (vx));

	for (i=0; i<vx; i++) {
		result.ptr->vr(i) = x.ptr->vr(i);
		result.ptr->ex(i) = n*x.ptr->ex(i);
	};

	result.ptr->vars = vx;

	return result;
}

monomial pow (const var& v, const int& n)
	{return pow (monomial(v),n);};



// Comparison of two monomials. Return <0 if the first one is smaller, >0 if
// the second one is smaller, and 0 if they're equal. Comparison is done
// using the reverse lexicographical order.

int cmp (const monomial& x, const monomial& y)
{
	short i = 0;
	int j;
	short vx = x.ptr->vars;
	short vy = y.ptr->vars;

	while (i<vx || i<vy) {
		if (i>=vx) return -1;
		if (i>=vy) return 1;
		j = y.ptr->vr(i)-x.ptr->vr(i);
		if (j) return j;
		j = x.ptr->ex(i)-y.ptr->ex(i);
		if (j) return j;
		i++;
	}

	return 0;
}

bool operator== (const monomial& x, const monomial& y) {
	if (!x.ptr || !y.ptr) return 0;
	return !cmp(x,y);
};
bool operator!= (const monomial& x, const monomial& y) {
	if (!x.ptr || !y.ptr) return 0;
	return cmp(x,y);
};
bool operator< (const monomial& x, const monomial& y) {
	if (!x.ptr || !y.ptr) return 0;
	return cmp(x,y)<0;
};
bool operator<= (const monomial& x, const monomial& y) {
	if (!x.ptr || !y.ptr) return 0;
	return cmp(x,y)<=0;
};
bool operator> (const monomial& x, const monomial& y) {
	if (!x.ptr || !y.ptr) return 0;
	return cmp(x,y)>0;
};
bool operator>= (const monomial& x, const monomial& y) {
	if (!x.ptr || !y.ptr) return 0;
	return cmp(x,y)>=0;
};
bool operator== (const monomial& x, const var& v)
	{return x==monomial(v);};
bool operator== (const var& v, const monomial& x)
	{return x==monomial(v);};
bool operator!= (const monomial& x, const var& v)
	{return x!=monomial(v);};
bool operator!= (const var& v, const monomial& x)
	{return x!=monomial(v);};



// Casting

bool is_undefined (const monomial& x) {return !x.ptr;};
bool is_var (const monomial& x) {
        return x.ptr && x.ptr->vars==1 && x.ptr->ex(0)==1;
};
var to_var (const monomial& x) {
	var v;
	setnum (v,x.ptr->vr(0));
	return v;
};
bool is_one (const monomial& x) {
	return x.ptr && !x.ptr->vars;
};



// Degree

int deg (const monomial& x, int varnum) {
	if (!x.ptr) return 0;
	for (int i=0; i<x.ptr->vars; i++)
		if (x.ptr->vr(i)==varnum) return x.ptr->ex(i);
	return 0;
};
int deg (const monomial& x, var v) {return deg (x,getnum (v));};



// Write a monomial to an output stream

ostream& operator<< (ostream& os, const monomial& x)
{
	int i,n;
	var z;

	if (!x.ptr) cout << "UNDEFINED";
	else if (x.ptr->vars) for (i=0; i<x.ptr->vars; i++) {
		setnum (z, x.ptr->vr(i));
		os << z;
		n = x.ptr->ex(i);
		if (n>1) os << n;
	} else os << "1";

	return os;
}



// Read a monomial from an input stream

istream& operator>> (istream& is, monomial& x)
{
	x = 1;
	var v;
	int n;
	char c;

	do {
		cin >> v;
		if (!is) goto BAD;
		n = 1;
		if (!is.eof()) {
			c = is.get();
			if (!is) goto BAD;
			is.unget ();
			if (c>='0' && c<='9') {
				cin >> n;
				if (!is) goto BAD;
			}
		}
		x *= pow(v,n);
		if (is.eof()) break;
		c = is.get();
		if (!is) goto BAD;
		is.unget ();
	} while ((c>='A' && c<='Z') || (c>='a' && c<='z'));

	return is;

BAD:	is.clear ( ios::badbit | is.rdstate () );
	return is;
}



// Database functions. We store the number of variables first, and then
// the variable number and the exponent for each variable.

void set_key (database& db, const monomial& x)
{
	if (!x.ptr) {
		set_key (db, -1);
		return;
	}

	set_key (db, x.ptr->vars);
	for (int i = 0; i<x.ptr->vars; i++) {
		set_key (db, x.ptr->vr(i));
		set_key (db, x.ptr->ex(i));
	}
}



void set_value (database& db, const monomial& x)
{
	if (!x.ptr) {
		set_value (db, -1);
		return;
	}

	set_value (db, x.ptr->vars);
	for (int i = 0; i<x.ptr->vars; i++) {
		set_value (db, x.ptr->vr(i));
		set_value (db, x.ptr->ex(i));
	}
}



bool get_value (database& db, monomial& x)
{
	int n;
	x = monomial ();

	if ( ! get_value (db,n) ) return 0;
	if (n<0) return 1;

	assert ( x.ptr = new monomial_data (n) );
	for (int i = 0; i<n; i++) {
		if ( ! get_value (db, x.ptr->vr(i)) ) return 0;
		if ( ! get_value (db, x.ptr->ex(i)) ) return 0;
	}

	return 1;
}

