/*
 * poly.cc - a C++ class for polynomial arithmetic
 *
 */

#include <assert.h>

#include "poly.h"

/*
 * If we include a file defining POLY_DEBUG, debugging output is printed
 * whenever a poly_data object is created or deleted. The last message
 * printed at program exit should be that the total poly_data records are 0.
 *
 */

#include "config.h"



void poly_data::poly_debug (int d)
{
	#ifdef POLY_DEBUG

	static int totalpolydata = 0;

	totalpolydata += d;

	cout << "*poly* total poly_data records " << totalpolydata << endl;

	assert (totalpolydata>=0);

	#endif
}



// Addition and subtraction. code is 0 for addition and 1 for subtraction.
// Basically we get the "union" of the two input polynomials. But we have
// to take care that the output polynomial is sorted.

poly paddsub (const poly& x, const poly& y, int code)
{
	poly result;
	poly_data *ptr1 = x.ptr;
	poly_data *ptr2 = y.ptr;
	poly_data *ptr1n;
	poly_data *ptr2n;
	poly_data *ptr3;
	rational z;
	int cp;

	if (!ptr1 || !ptr2) return result;

	if (is_zero (y)) return x;
	if (is_zero (x)) if (code) return -y;
		else return y;

	assert (ptr3 = result.ptr = new poly_data);

	while (ptr1->next || ptr2->next) {
		ptr1n = ptr1->next;
		ptr2n = ptr2->next;
		if (ptr1n && ptr2n) cp = cmp (ptr1n->mon, ptr2n->mon);
		if (!ptr2n || (ptr1n && cp>0)) {
			assert (ptr3 = ptr3->next = new poly_data);
			ptr1 = ptr1n;
			ptr3->cof = ptr1->cof;
			ptr3->mon = ptr1->mon;
			continue;
		};
		if (!ptr1n || (ptr2n && cp<0)) {
			assert (ptr3 = ptr3->next = new poly_data);
			ptr2 = ptr2n;
			if (code) ptr3->cof = -ptr2->cof;
			else ptr3->cof = ptr2->cof;
			ptr3->mon = ptr2->mon;
			continue;
		};
		ptr1 = ptr1n;
		ptr2 = ptr2n;
		if (code) z = ptr1->cof - ptr2->cof;
		else z = ptr1->cof + ptr2->cof;
		if (!is_zero(z)) {
			assert (ptr3 = ptr3->next = new poly_data);
			ptr3->cof = z;
			ptr3->mon = ptr1->mon;
		};
	}

	return result;
}

#define PADDSUB(TYPE1,TYPE2)				\
poly operator+ (const TYPE1& x, const TYPE2& y)	\
	{return paddsub (x,y,0);};			\
poly operator- (const TYPE1& x, const TYPE2& y)	\
	{return paddsub (x,y,1);};
PADDSUB (poly,poly);
PADDSUB (poly,monomial);
PADDSUB (monomial,poly);
PADDSUB (monomial,monomial);
PADDSUB (poly,var);
PADDSUB (var,poly);
PADDSUB (var,var);
PADDSUB (monomial,var);
PADDSUB (var,monomial);
PADDSUB (poly,rational);
PADDSUB (monomial,rational);
PADDSUB (var,rational);
PADDSUB (rational,poly);
PADDSUB (rational,monomial);
PADDSUB (rational,var);
PADDSUB (poly,int);
PADDSUB (monomial,int);
PADDSUB (var,int);
PADDSUB (int,poly);
PADDSUB (int,monomial);
PADDSUB (int,var);



// Addition and subtraction in += / -= version. We provide a function for
// this that avoids frequent memory (de-)allocation by re-using the "old"
// monomials if possible.

poly& poly::paddsub2 (const poly& x, int code)
{
	int cp;

	if (!ptr) return *this;
	if (!x.ptr) return *this = poly ();

	// If the reference count is bigger than 1, we cannot overwrite the
	// old polynomial. Use the standard +/- functions then.
	if ( ptr->cnt()>1 ) return *this = paddsub (*this, x, code);

	// The reference count is 1. We can overwrite the old polynomial.
	// Do the addition/subtraction by inserting the monomials of x into
	// the current polynomial.
	poly_data *ptr1 = x.ptr;
	poly_data *ptr2 = ptr;
	poly_data *p;

	while (ptr1 = ptr1->next) {
		// Insert the monomial in ptr1 into the polynomial. First,
		// find the correct position.
		cp = 1;
		while (ptr2->next) {
			cp = cmp (ptr1->mon, ptr2->next->mon);
			if (cp<0) {
				ptr2 = ptr2->next;
				continue;
			}
			break;
		}

		if (cp) {
			// The monomial doesn't exist yet in the
			// current polynomial. Insert it.
			p = ptr2->next;
			assert (ptr2 = ptr2->next = new poly_data);
			if (code) ptr2->cof = - ptr1->cof;
			else ptr2->cof = ptr1->cof;
			ptr2->mon = ptr1->mon;
			ptr2->next = p;
		} else {
			// The monomial exists already. Add/subtract
			// the coefficients.
			if (code) ptr2->next->cof -= ptr1->cof;
			else ptr2->next->cof += ptr1->cof;
			if (is_zero (ptr2->next->cof)) {
				// New coefficient is 0. Delete entry.
				p = ptr2->next->next;
				delete ptr2->next;
				ptr2->next = p;
			}
		}
	}

	return *this;
}



// Multiplication and division with rationals and monomials. The rational
// and monomial arguments may be UNDEFINED, in which case this multiplication
// is left out.

poly pmul (const poly& x, const monomial& y, const rational& z)
{
	if (!x.ptr) return x;
	if (is_zero(z)) return 0;

	poly result = 0;

	poly_data *ptr1 = x.ptr;
	poly_data *ptr2 = result.ptr;

	while (ptr1 = ptr1->next) {
		assert (ptr2 = ptr2->next = new poly_data);
		if (is_undefined (y)) ptr2->mon = ptr1->mon;
		else ptr2->mon = ptr1->mon * y;
		if (is_undefined (z)) ptr2->cof = ptr1->cof;
		else ptr2->cof = ptr1->cof * z;
	}

	return result;
}



poly pdiv (const poly& x, const monomial& y, const rational& z)
{
	if (!x.ptr) return x;
	if (is_zero(z)) return poly ();

	poly result = 0;

	poly_data *ptr1 = x.ptr;
	poly_data *ptr2 = result.ptr;

	while (ptr1 = ptr1->next) {
		assert (ptr2 = ptr2->next = new poly_data);
		if (is_undefined (y)) ptr2->mon = ptr1->mon;
		else if (is_undefined (ptr2->mon = ptr1->mon / y))
			return poly ();
		if (is_undefined (z)) ptr2->cof = ptr1->cof;
		else ptr2->cof = ptr1->cof / z;
	}

	return result;
}

#define PMULDIVM(TYPE1,TYPE2)			\
poly operator* (const TYPE1& x, const TYPE2& y)	\
	{return pmul (x,y,rational ());};	\
poly operator/ (const TYPE1& x, const TYPE2& y)	\
	{return pdiv (x,y,rational ());};	\
poly operator* (const TYPE2& y, const TYPE1& x)	\
	{return pmul (x,y,rational ());};
#define PMULDIVR(TYPE1,TYPE2)			\
poly operator* (const TYPE1& x, const TYPE2& y)	\
	{return pmul (x,monomial (),y);};	\
poly operator/ (const TYPE1& x, const TYPE2& y)	\
	{return pdiv (x,monomial (),y);};	\
poly operator* (const TYPE2& y, const TYPE1& x)	\
	{return pmul (x,monomial (),y);};
PMULDIVM (poly,monomial);
PMULDIVM (poly,var);
PMULDIVR (poly,rational);
PMULDIVR (poly,int);
PMULDIVR (monomial,rational);
PMULDIVR (monomial,int);
PMULDIVR (var,rational);
PMULDIVR (var,int);



poly& poly::pmul2 (const monomial& y, const rational& z)
{
	if (!ptr) return *this;
	if (is_zero(z)) return *this = poly (0);

	if ( ptr->cnt() > 1 ) return *this = pmul (*this,y,z);

	poly_data *ptr1 = ptr;

	while (ptr1 = ptr1->next) {
		if (!is_undefined (y)) ptr1->mon *= y;
		if (!is_undefined (z)) ptr1->cof *= z;
	}

	return *this;
}



poly& poly::pdiv2 (const monomial& y, const rational& z)
{
	if (!ptr) return *this;
	if (is_zero(z)) return *this = poly ();

	if ( ptr->cnt() > 1 ) return *this = pdiv (*this,y,z);

	poly_data *ptr1 = ptr;

	while (ptr1 = ptr1->next) {
		if (!is_undefined (y)) ptr1->mon /= y;
		if (!is_undefined (z)) ptr1->cof /= z;
	}

	return *this;
}



// Multiplication of polynomials

poly operator* (const poly& x, const poly& y)
{
	poly result;
	if (!x.ptr || !y.ptr) return result;

	result = 0;
	if (is_zero(x) || is_zero(y)) return result;

	poly_data *ptr1 = x.ptr;
	while (ptr1 = ptr1->next) result += pmul (y, ptr1->mon, ptr1->cof);

	return result;
}



// Division of polynomials. Return UNDEFINED if the division doesn't go.

poly operator/ (const poly& x, const poly& y)
{
	poly result;
	if (!x.ptr || !y.ptr || is_zero (y)) return result;

	result = 0;
	if (is_zero (x)) return result;

	poly p = x;
	monomial m;
	rational r;

	while (!is_zero(p)) {
		if (is_undefined (m = p.ptr->next->mon / y.ptr->next->mon))
			return poly ();
		r = p.ptr->next->cof / y.ptr->next->cof;
		result += r*m;
		p -= pmul (y,m,r);
	};

	return result;
}



// Exponentiation

poly pow (const poly& x, const int& n)
{
	int i;

	if (!x.ptr) return x;

	if (is_rational(x)) return pow (to_rational(x),n);
	if (n<0) return poly ();
	if (!n) return 0;

	if (!x.ptr->next->next) // only one summand -> easy!
		return pow(x.ptr->next->cof,n)*pow(x.ptr->next->mon,n);

	poly p = x;
	for (i=2; i<=n; i++) p *= x;

	return p;
}



// Signs

poly operator+ (const poly& x) {return x;};
poly operator+ (const monomial& x) {return poly(x);};
poly operator+ (const var& x) {return poly(x);};
poly operator- (const poly& x) {return (-1)*x;};
poly operator- (const monomial& x) {return (-1)*poly(x);};
poly operator- (const var& x) {return (-1)*poly(x);};



// Comparisons of two polynomials. Return -1 if one of them is UNDEFINED,
// 1 if they are equal, and 0 otherwise.

int pcmp (const poly& x, const poly& y)
{
	if (!x.ptr || !y.ptr) return -1;

	poly_data *ptr1 = x.ptr;
	poly_data *ptr2 = y.ptr;

	while (ptr1->next || ptr2->next) {
		if (!ptr1->next || !ptr2->next) return 0;
		ptr1 = ptr1->next;
		ptr2 = ptr2->next;
		if (ptr1->cof != ptr2->cof) return 0;
		if (ptr2->mon != ptr2->mon) return 0;
	}

	return 1;
}

bool operator== (const poly& x, const poly& y)
	{return pcmp (x,y)==1;};
bool operator!= (const poly& x, const poly& y)
	{return !pcmp (x,y);};
#define PCMP(TYPE,COMP,CONV)				\
bool operator== (const poly& x, const TYPE& y) {	\
	if (!COMP (x)) return 0;			\
	return CONV (x)==y;				\
};							\
bool operator== (const TYPE& y, const poly& x) {	\
	if (!COMP (x)) return 0;			\
	return CONV (x)==y;				\
};							\
bool operator!= (const poly& x, const TYPE& y) {	\
	if (!COMP (x)) return 1;			\
	return CONV (x)!=y;				\
};							\
bool operator!= (const TYPE& y, const poly& x) {	\
	if (!COMP (x)) return 1;			\
	return CONV (x)!=y;				\
};
PCMP (monomial,is_monomial,to_monomial);
PCMP (var,is_var,to_var);
PCMP (rational,is_rational,to_rational);
PCMP (int,is_int,to_int);



// Casting

bool is_undefined (const poly& x) {
	return !x.ptr;
};
bool is_zero (const poly& x) {
	return x.ptr && !x.ptr->next;
};
bool is_monomial (const poly& x) {
	return x.ptr && x.ptr->next && ! x.ptr->next->next
		&& x.ptr->next->cof==1;
};
monomial to_monomial (const poly& x) {
	if (!is_monomial (x)) return monomial ();
	return x.ptr->next->mon;
};
bool is_var (const poly& x) {
	return is_monomial (x) && is_var (x.ptr->next->mon);
};
var to_var (const poly& x) {
	return to_var (x.ptr->next->mon);
};
bool is_rational (const poly& x) {
	if (is_zero (x)) return 1;
	return x.ptr && x.ptr->next && ! x.ptr->next->next
		&& x.ptr->next->mon==monomial (1);
};
rational to_rational (const poly& x)
{
	if (!is_rational (x)) return rational ();
	if (is_zero(x)) return 0;
	return x.ptr->next->cof;
};
bool is_int (const poly& x) {
	if ( !is_rational (x) ) return 0;
	if ( is_zero (x) ) return 1;
	return is_int (x.ptr->next->cof);
};
int to_int (const poly& x) {
        if (!is_int (x) || is_zero(x)) return 0;
	return to_int (x.ptr->next->cof);
};



// Conversion for division

poly divconv (var x) {return poly(x);};
poly divconv (monomial x) {return poly(x);};
poly divconv (poly x) {return x;};



// Degrees and coefficients

int deg (const poly& x, int varnum)
{
	if (!x.ptr) return 0;

	int d = 0, e;
	poly_data *ptr = x.ptr;

	while (ptr = ptr->next) {
		e = deg (ptr->mon, varnum);
		d = (d>e ? d : e);
	}

	return d;
}



poly coeff (const poly& x, int varnum, int d)
{
	if (!x.ptr) return x;

	poly result = 0;
	poly_data *ptr = x.ptr;

	while (ptr = ptr->next)
		if ( deg (ptr->mon, varnum) == d)
			result += ptr->cof * ptr->mon;

	var v;
	setnum (v,varnum);

	return result/pow(v,d);
}



int deg (const poly& x, const var& v)
	{return deg (x,getnum (v));};
poly coeff (const poly& x, const var& v, int d)
	{return coeff (x,getnum (v),d);};
poly coeff (const monomial& x, const var& v, const int& d)
        {return coeff (poly(x),v,d);};



// Substitution

poly subs (const poly& x, const var& v, const poly& y)
{
	if (!x.ptr || !y.ptr) return poly ();

	int d = deg (x,v);
	poly result = 0;

	for (int i=d; i>=0; i--) result = (result*y)+coeff (x,v,i);

	return result;
}



// Write a polynomial to an output stream

ostream& operator<< (ostream& os, const poly& x)
{
	int sign = 0;
	poly_data *p;

	if  (!x.ptr) os << "UNDEFINED";
	else if (!x.ptr->next) os << "0";
	else {
		p = x.ptr;
		while (p = p->next) {
			if (sign++ && p->cof>0) os << "+";
			if (is_one (p->mon)) os << p->cof;
			else {
				if (p->cof != 1) {
					if (p->cof == -1) os << "-";
					else os << p->cof;
				}
				os << p->mon;
			}
		};
	}

	return os;
}



// Read a polynomial from an input stream

istream& operator>> (istream& is, poly& x)
{
	char c;
	int sign;
	rational z;
	monomial m;
	int i;

	x = 0;

	do {
		// Read sign
		sign = 1;
		if (is.eof ()) goto BAD;
		c = is.get ();
		while (c==' ') {
			if (!is) goto BAD;
			c = is.get ();
		}
		if (c=='-') sign = -1;
		else if (c!='+') is.unget ();

		// Read coefficient
		i = 0;
		z = 1;
		if (is.eof ()) goto BAD;
		c = is.get ();
		if (!is) goto BAD;
		is.unget ();
		if (c>='0' && c<='9') {
			cin >> z;
			if (!is) goto BAD;
			i = 1;
		}

		// Read monomial
		m = 1;
		if (!is.eof ()) {
			c = is.get ();
			if (!is) goto BAD;
			is.unget ();
			if ((c>='A' && c<='z') || (c>='a' && c<='z')) {
				cin >> m;
				if (!is) goto BAD;
				i = 1;
			}
		}

		if (!i) goto BAD; // need a number or a monomial

		x += sign*z*m;

		// is there another monomial?
		if (is.eof ()) continue;
		c = is.get ();
		if (!is) goto BAD;
		is.unget ();
	} while (c=='+' || c=='-');

	return is;

BAD:	is.clear ( ios::badbit | is.rdstate() );
	return is;
}



// Database functions. For every term, we store the coefficient and the
// monomial. This stops if
//  1) a coefficient is 0 (in which case there is no monomial stored any more)
//  2) a monomial is constant (in which case there is no further coefficient
//     stored). This is possible because polynomials are always sorted; so
//     a constant term must be the last one.
// The undefined polynomial is stored by having the first coefficient undefined
// (and no further entries). The 0 polynomial just has a coefficient 0.

void set_key (database& db, const poly& x)
{
	if (!x.ptr) {
		set_key (db, rational());
		return;
	}

	poly_data *ptr = x.ptr;
	while (ptr = ptr->next) {
		set_key (db, ptr->cof);
		set_key (db, ptr->mon);
		if ( is_one (ptr->mon) ) return;
	}

	set_key (db, rational(0));
}



void set_value (database& db, const poly& x)
{
	if (!x.ptr) {
		set_value (db, rational());
		return;
	}

	poly_data *ptr = x.ptr;
	while (ptr = ptr->next) {
		set_value (db, ptr->cof);
		set_value (db, ptr->mon);
		if ( is_one (ptr->mon) ) return;
	}

	set_value (db, rational(0));
}



bool get_value (database& db, poly& x)
{
	x = 0;
	rational r;
	poly_data *ptr = x.ptr;

	do {
		if ( ! get_value (db, r) ) return 0;
		if ( is_undefined (r) ) {
			x = poly ();
			return 1;
		}
		if ( is_zero (r) ) return 1;
		assert (ptr = ptr->next = new poly_data);
		ptr->cof = r;
		if ( ! get_value (db, ptr->mon) ) return 0;
	} while ( ! is_one (ptr->mon) );

	return 1;
}

